"""Data module"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/05_hf.data.base.ipynb.

# %% auto 0
__all__ = ['Data']

# %% ../../../nbs/05_hf.data.base.ipynb 3
"""
Data module
"""

from .tokens import Tokens


class Data:
    """
    Base data tokenization class.
    """

    def __init__(self, tokenizer, columns, maxlength):
        """
        Creates new base instance for tokenizing data.

        Args:
            tokenizer: model tokenizer
            columns: column names
            maxlength: maximum sequence length
        """

        self.tokenizer = tokenizer
        self.columns = columns
        self.maxlength = maxlength

    def __call__(self, train, validation, workers):
        """
        Tokenizes training and validation data and returns processed datasets.

        Args:
            train: training data
            validation: validation data
            workers: number of concurrent tokenizers when processing datasets, only main process used when set to None

        Returns:
            (train, validation)
        """

        return (self.prepare(train, self.process, workers), self.prepare(validation, self.process, workers) if validation else None)

    def prepare(self, data, fn, workers):
        """
        Prepares and tokenizes data for training.

        Args:
            data: input data
            fn: tokenize processing function to apply
            workers: number of concurrent tokenizers when processing datasets, only main process used when set to None

        Returns:
            tokens
        """

        if hasattr(data, "map"):
            # Hugging Face dataset
            tokens = data.map(fn, batched=True, num_proc=workers, remove_columns=data.column_names)
        else:
            # Re-orient data into columns for efficient batch tokenization
            columns = {}
            if hasattr(data, "columns"):
                # Polars/pandas DataFrame
                for column in data.columns:
                    columns[column] = list(data[column])
            else:
                # Iterable dicts
                for row in data:
                    for column in row.keys():
                        if column not in columns:
                            columns[column] = []

                        columns[column].append(row[column])

            # Process column-oriented data
            tokens = Tokens(fn(columns))

        return tokens

    def labels(self, data):
        """
        Extracts a list of unique labels from data.

        Args:
            data: input data

        Returns:
            list of unique labels
        """

        # Last column is label
        column = self.columns[-1]

        # Return length of labels if it's an array
        length = self.length(data[column][0] if hasattr(data, "columns") else data[0][column])
        if length:
            return length

        if hasattr(data, "map"):
            # Hugging Face dataset
            labels = sorted(data.unique(self.columns[-1]))
        elif hasattr(data, "columns"):
            # Polars/pandas DataFrame
            labels = sorted(data[self.columns[-1]].unique())
        else:
            # Iterable dicts
            labels = sorted({row[self.columns[-1]] for row in data})

        # Labels are single numeric values per entry
        #   - Consider a regression task if at least one label isn't an integer
        #   - Otherwise use number of labels for a classification task
        return 1 if [x for x in labels if float(x) != int(x)] else len(labels)

    def process(self, data):
        """
        Tokenizes batch of input data

        Args:
            data: input data batch

        Returns:
            tokenized data
        """

        return data

    def length(self, value):
        """
        Returns the length of value if value has a len function defined. Otherwise,
        None is returned.

        Args:
            value: value to check

        Returns:
            length of value if available, otherwise returns None
        """

        return len(value) if hasattr(value, "__len__") else None




"""Sequences module"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/05_hf.data.sequences.ipynb.

# %% auto 0
__all__ = ['Sequences']

# %% ../../../nbs/05_hf.data.sequences.ipynb 3
"""
Sequences module
"""

from .base import Data


class Sequences(Data):
    """
    Tokenizes sequence-sequence datasets as input for training sequence-sequence models
    """

    def __init__(self, tokenizer, columns, maxlength, prefix):
        """
        Creates a new instance for tokenizing Sequences training data.

        Args:
            tokenizer: model tokenizer
            columns: tuple of columns to use for text/label
            maxlength: maximum sequence length
            prefix: source prefix
        """

        super().__init__(tokenizer, columns, maxlength)

        # Standardize columns
        if not self.columns:
            self.columns = ("source", "target")

        # Save source prefix
        self.prefix = prefix

    def process(self, data):
        # Column keys
        source, target = self.columns

        # Tokenize source
        source = [self.prefix + x if self.prefix else x for x in data[source]]
        inputs = self.tokenizer(source, max_length=self.maxlength, padding=False, truncation=True)

        # Tokenize target
        with self.tokenizer.as_target_tokenizer():
            targets = self.tokenizer(data[target], max_length=self.maxlength, padding=False, truncation=True)

        # Combine inputs
        inputs["labels"] = targets["input_ids"]

        return inputs




"""CLS module"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/05_hf.models.pooling.cls.ipynb.

# %% auto 0
__all__ = ['ClsPooling']

# %% ../../../../nbs/05_hf.models.pooling.cls.ipynb 3
"""
CLS module
"""

from .base import Pooling


class ClsPooling(Pooling):
    """
    Builds CLS pooled vectors using outputs from a transformers model.
    """

    def forward(self, **inputs):
        """
        Runs CLS pooling on token embeddings.

        Args:
            inputs: model inputs

        Returns:
            CLS pooled embeddings using output token embeddings (i.e. last hidden state)
        """

        # Run through transformers model
        tokens = super().forward(**inputs)

        # CLS token pooling
        return tokens[:, 0]




"""Pooling module"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../nbs/05_hf.models.pooling.base.ipynb.

# %% auto 0
__all__ = ['Pooling']

# %% ../../../../nbs/05_hf.models.pooling.base.ipynb 3
"""
Pooling module
"""

import numpy as np
import torch

from torch import nn

from ..models import Models


class Pooling(nn.Module):
    """
    Builds pooled vectors usings outputs from a transformers model.
    """

    def __init__(self, path, device, tokenizer=None, maxlength=None, modelargs=None):
        """
        Creates a new Pooling model.

        Args:
            path: path to model, accepts Hugging Face model hub id or local path
            device: tensor device id
            tokenizer: optional path to tokenizer
            maxlength: max sequence length
            modelargs: additional model arguments
        """

        super().__init__()

        self.model = Models.load(path, modelargs=modelargs)
        self.tokenizer = Models.tokenizer(tokenizer if tokenizer else path)
        self.device = Models.device(device)

        # Detect unbounded tokenizer typically found in older models
        Models.checklength(self.model, self.tokenizer)

        # Set max length
        self.maxlength = maxlength if maxlength else self.tokenizer.model_max_length if self.tokenizer.model_max_length != int(1e30) else None

        # Move to device
        self.to(self.device)

    def encode(self, documents, batch=32):
        """
        Builds an array of pooled embeddings for documents.

        Args:
            documents: list of documents used to build embeddings
            batch: model batch size

        Returns:
            pooled embeddings
        """

        # Split documents into batches and process
        results = []

        # Sort document indices from largest to smallest to enable efficient batching
        # This performance tweak matches logic in sentence-transformers
        lengths = np.argsort([-len(x) if x else 0 for x in documents])
        documents = [documents[x] for x in lengths]

        for chunk in self.chunk(documents, batch):
            # Tokenize input
            inputs = self.tokenizer(chunk, padding=True, truncation="longest_first", return_tensors="pt", max_length=self.maxlength)

            # Move inputs to device
            inputs = inputs.to(self.device)

            # Run inputs through model
            with torch.no_grad():
                outputs = self.forward(**inputs)

            # Add batch result
            results.extend(outputs.cpu().numpy())

        # Restore original order and return array
        return np.asarray([results[x] for x in np.argsort(lengths)])

    def chunk(self, texts, size):
        """
        Splits texts into separate batch sizes specified by size.

        Args:
            texts: text elements
            size: batch size

        Returns:
            list of evenly sized batches with the last batch having the remaining elements
        """

        return [texts[x : x + size] for x in range(0, len(texts), size)]

    def forward(self, **inputs):
        """
        Runs inputs through transformers model and returns outputs.

        Args:
            inputs: model inputs

        Returns:
            model outputs
        """

        return self.model(**inputs)[0]




import os
import yaml
import streamlit as st
from pathlib import Path
import mimetypes
from onprem import utils as U
from onprem.app.utils import load_llm

# https://github.com/VikParuchuri/marker/issues/442#issuecomment-2636393925
import torch
torch.classes.__path__ = [os.path.join(torch.__path__[0], torch.classes.__file__)]

DATADIR = U.get_datadir()
DEFAULT_YAML_FNAME = "config.yml"
DEFAULT_WEBAPP_DIR = U.get_webapp_dir()
DEFAULT_YAML_FPATH = os.path.join(DEFAULT_WEBAPP_DIR, DEFAULT_YAML_FNAME)
DEFAULT_PROMPT = "List three cute names for a cat."
DEFAULT_YAML = """
llm:
  # model url (or model file name if previously downloaded)
  # if changing, be sure to update the prompt_template variable below
  model_url: https://huggingface.co/TheBloke/zephyr-7B-beta-GGUF/resolve/main/zephyr-7b-beta.Q4_K_M.gguf
  # number of layers offloaded to GPU
  n_gpu_layers: -1
  # type of vector store
  # ("dual" means both Chroma semantic searches and conventional keyword searches are supported)
  store_type: dual
  # path to vector db folder
  vectordb_path: {webapp_dir}/vectordb
  # path to model download folder
  model_download_path: {models_dir}
  # number of source documents used by LLM.ask and LLM.chat
  rag_num_source_docs: 6
  # minimum similarity score for source to be considered by LLM.ask/LLM.chat
  rag_score_threshold: 0.0
  # verbosity of Llama.cpp
  verbose: TRUE
  # additional parameters added in the "llm" YAML section will be fed directly to LlamaCpp (e.g., temperature)
  #temperature: 0.0
  # max_tokens: 2048
prompt:
  # The default prompt_template is specifically for the Zephyr-7B model.
  # It will need to be changed if you change the model_url above.
  prompt_template: <|system|>\\n</s>\\n<|user|>\\nPROMPT_VARIABLE</s>\\n<|assistant|>
ui:
  # title of application
  title: OnPrem.LLM
  # subtitle in "Talk to Your Documents" screen
  rag_title:
  # path to markdown file with contents that will be inserted below rag_title
  rag_text_path:
  # path to folder containing raw documents (i.e., absolute path of folder you supplied to LLM.ingest)
  rag_source_path: {webapp_dir}/documents
  # base url (leave blank unless you're running your own separate web server to serve source documents)
  rag_base_url:
  # whether to show the Manage page in the sidebar (TRUE or FALSE)
  show_manage: TRUE
"""


def write_default_yaml():
    """
    write default webapp.yml
    """
    yaml_content = DEFAULT_YAML.format(
        webapp_dir=U.get_webapp_dir(),
        models_dir=U.get_models_dir()
    ).strip()
    yaml_content = yaml_content.replace('PROMPT_VARIABLE', '{prompt}')
    with open(DEFAULT_YAML_FPATH, "w") as f:
        f.write(yaml_content)
    return


def read_config():
    """
    Read config file.  Returns a dictionary of the configuration and a boolean indicating whether or not a new config was created.
    """
    exists = os.path.exists(DEFAULT_YAML_FPATH)
    if not exists:
        write_default_yaml()
    with open(DEFAULT_YAML_FPATH, "r") as stream:
        cfg = yaml.safe_load(stream)
    return cfg, not exists


def is_txt(fpath):
    try:
        result = mimetypes.guess_type(fpath)
        return result[0] and result[0].startswith("text/")
    except:
        return False


def main():
    """
    Main entry point for the Streamlit application
    """
    # Page setup
    cfg, cfg_was_created = read_config()

    TITLE = cfg.get("ui", {}).get("title", "OnPrem.LLM")
    st.set_page_config(
        page_title=TITLE, 
        page_icon="🐍", 
        layout="wide",
        initial_sidebar_state="expanded"
    )
    
    # First run configuration warning
    if cfg_was_created:
        st.warning(
            f"No {DEFAULT_YAML_FNAME} file was found in {DATADIR}, so a default one was created for you. Please edit as necessary."
        )
    
    # Home page content
    st.header("Welcome to OnPrem.LLM")
    
    # Load LLM to get model name
    ## 2025-04-09: commenting this out, a streamlit is reloading the again model in other pages
    #llm = load_llm()
    #model_name = llm.model_name
    
    # Main content
    st.markdown("""
    This application allows you to interact with an LLM in multiple ways:
    
    1. **Use Prompts to Solve Problems**: Submit prompts directly to the LLM model
    2. **Talk to Your Documents**: Ask questions about your documents using RAG technology
    3. **Document Analysis**: Apply custom prompts to document chunks and export results
    4. **Search Documents**: Search through your indexed documents using keywords or semantic search
    5. **Manage**: Upload documents, manage folders, and configure the application settings
    
    Use the sidebar to navigate between these different features.
    """)
    
    ## see 2025-04-09 comment above
    # Model information
    #st.subheader("Current Model Information")
    #st.markdown(f"**Model**: {model_name}")
    
    # Quick links
    st.subheader("Quick Links")
    col1, col2, col3, col4 = st.columns(4)
    
    with col1:
        if st.button("💬 Use Prompts", use_container_width=True):
            st.switch_page("pages/1_Prompts.py")
    
    with col2:
        if st.button("📄 Talk to Documents", use_container_width=True):
            st.switch_page("pages/2_Document_QA.py")
    
    with col3:
        if st.button("📊 Document Analysis", use_container_width=True):
            st.switch_page("pages/3_Document_Analysis.py")
    
    with col4:
        if st.button("🔍 Search Documents", use_container_width=True):
            st.switch_page("pages/4_Document_Search.py")
    
    # Additional information
    st.markdown("---")
    st.markdown("Built with [OnPrem](https://github.com/amaiya/onprem)")


if __name__ == "__main__":
    main()



"""functionality for text extraction and document ingestion into a vector database for question-answering and other tasks"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_ingest.base.ipynb.

# %% auto 0
__all__ = ['logger', 'DEFAULT_CHUNK_SIZE', 'DEFAULT_CHUNK_OVERLAP', 'TABLE_CHUNK_SIZE', 'CHROMA_MAX', 'PDFOCR', 'PDFMD', 'PDF',
           'PDF_EXTS', 'OCR_CHAR_THRESH', 'LOADER_MAPPING', 'MyElmLoader', 'MyUnstructuredPDFLoader',
           'PDF2MarkdownLoader', 'load_single_document', 'load_documents', 'process_folder', 'chunk_documents',
           'does_vectorstore_exist', 'batchify_chunks', 'VectorStore']

# %% ../../nbs/01_ingest.base.ipynb 3
from ..llm.helpers import summarize_tables, extract_title
from ..utils import batch_list, filtered_generator
from . import helpers

from langchain_core.documents import Document
from langchain_huggingface import HuggingFaceEmbeddings
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_text_splitters.base import Language
from langchain_community.document_loaders import (
    CSVLoader,
    EverNoteLoader,
    TextLoader,
    PyMuPDFLoader,
    UnstructuredPDFLoader,
    UnstructuredEmailLoader,
    UnstructuredEPubLoader,
    UnstructuredHTMLLoader,
    UnstructuredMarkdownLoader,
    UnstructuredODTLoader,
    UnstructuredPowerPointLoader,
    UnstructuredWordDocumentLoader,
    UnstructuredExcelLoader,
)


import os
import os.path
from typing import List, Optional, Callable
import multiprocessing
import functools
from tqdm import tqdm
import warnings
import logging
import sys

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger('OnPrem.LLM-ingest')

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 50
TABLE_CHUNK_SIZE = 2000
CHROMA_MAX = 41000

# %% ../../nbs/01_ingest.base.ipynb 4
class MyElmLoader(UnstructuredEmailLoader):
    """Wrapper to fallback to text/plain when default does not work"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        try:
            try:
                doc = UnstructuredEmailLoader.load(self)
            except ValueError as e:
                if "text/html content not found in email" in str(e):
                    # Try plain text
                    self.unstructured_kwargs["content_source"] = "text/plain"
                    doc = UnstructuredEmailLoader.load(self)
                else:
                    raise
        except Exception as e:
            raise Exception(f'{self.file_path} : {e}')

        return doc


class MyUnstructuredPDFLoader(UnstructuredPDFLoader):
    """Custom PDF Loader"""

    def load(self) -> List[Document]:
        """Wrapper UnstructuredPDFLoader"""
        try:
            docs = UnstructuredPDFLoader.load(self)
            if not docs:
                raise Exception('Document had no content. ')
            tables = [d.metadata['text_as_html'] for d in docs if d.metadata.get('text_as_html', None) is not None]
            texts = [d.page_content for d in docs if d.metadata.get('text_as_html', None) is None]

            page_content = '\n'.join(texts)
            docs = [helpers.create_document(page_content, self.file_path)]
            table_docs = [helpers.create_document(t, source=self.file_path, table=True) for t in tables]
            docs.extend(table_docs)
            return docs
        except Exception as e:
            # Add file_path to exception message
            raise Exception(f'{self.file_path} : {e}')


class _PyMuPDFLoader(PyMuPDFLoader):
    """Custom PyMUPDF Loader with optional support for inferring table structure"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        try:
            # PyMuPDFLoader complains when you add custom flags to text_kwargs,
            # so delete before loading
            infer_table_structure = self.parser.text_kwargs.get('infer_table_structure', False)
            if 'infer_table_structure' in self.parser.text_kwargs:
                del self.parser.text_kwargs['infer_table_structure']
            docs = PyMuPDFLoader.load(self)
            if infer_table_structure:
                docs = helpers.extract_tables(docs=docs)
            return docs
        except Exception as e:
            # Add file_path to exception message
            raise Exception(f'{self.file_path} : {e}')


class PDF2MarkdownLoader(_PyMuPDFLoader):
    """Custom PDF to Markdown Loader"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        import pymupdf4llm
        try:
            md_text = pymupdf4llm.to_markdown(self.file_path, show_progress=False)
            if not md_text.strip():
                raise Exception('Document had no content. ')
            doc = helpers.create_document(md_text, self.file_path, markdown=True)
            docs = [doc]
            if self.parser.text_kwargs.get('infer_table_structure', False):
                docs = helpers.extract_tables(docs=docs)
            return docs
        except Exception as e:
            # Add file_path to exception message
            raise Exception(f'{self.file_path} : {e}')




# %% ../../nbs/01_ingest.base.ipynb 5
# Map file extensions to document loaders and their arguments
PDFOCR = 'pdfOCR'
PDFMD = 'pdfMD'
PDF = 'pdf'
PDF_EXTS = [PDF, PDFOCR, PDFMD]
OCR_CHAR_THRESH = 32
LOADER_MAPPING = {
    "csv": (CSVLoader, {}),
    "doc": (UnstructuredWordDocumentLoader, {}),
    "docx": (UnstructuredWordDocumentLoader, {}),
    "xlsx": (UnstructuredExcelLoader, {}),
    "enex": (EverNoteLoader, {}),
    "eml": (MyElmLoader, {}),
    "epub": (UnstructuredEPubLoader, {}),
    "html": (UnstructuredHTMLLoader, {}),
    "htm": (UnstructuredHTMLLoader, {}),
    "md": (UnstructuredMarkdownLoader, {}),
    "odt": (UnstructuredODTLoader, {}),
    "ppt": (UnstructuredPowerPointLoader, {}),
    "pptx": (UnstructuredPowerPointLoader, {}),
    "txt": (TextLoader, {"autodetect_encoding": True}),
    "json": (TextLoader, {"autodetect_encoding": True}),
    PDF   : (_PyMuPDFLoader, {}),
    PDFMD: (PDF2MarkdownLoader, {}),
    PDFOCR: (MyUnstructuredPDFLoader, {"infer_table_structure":False, "mode":"elements", "strategy":"hi_res"}),
    # Add more mappings for other file extensions and loaders as needed
}

def _update_metadata(docs:List[Document], metadata:dict):
    """
    Update metadata in docs with supplied metadata dictionary
    """
    for doc in docs:
        doc.metadata.update(metadata)
    return docs

def _apply_text_callables(docs:List[Document], text_callables:dict):
    """
    Invokes text_callables on entire text of document.

    Returns a dictionary with values containing results from callables for each key
    """
    if not text_callables: return {}
        
    text = '\n\n'.join([d.page_content for d in docs])
    results = {}
    for k,v in text_callables.items():
        results[k] = v(text)
    return results

def _apply_file_callables(file_path:str, file_callables:dict):
    """
    Invokes file_callables on file path.

    Returns a dictionary with values containing results from callables for each key
    """
    if not file_callables: return {}
        
    if not os.path.exists(file_path):
        raise ValueError('file_path does not exist: {file_path}')
    
    results = {}
    for k,v in file_callables.items():
        results[k] = v(file_path)
    return results

    
def load_single_document(file_path: str, # path to file
                         pdf_unstructured:bool=False, # use unstructured for PDF extraction if True (will also OCR if necessary)
                         pdf_markdown:bool = False, # Convert PDFs to Markdown instead of plain text if True.
                         store_md5:bool=False, # Extract and store MD5 of document in metadata
                         store_mimetype:bool=False, # Guess and store mime type of document in metadata
                         store_file_dates:bool=False, # Extract snd store file dates in metadata
                         file_callables:Optional[dict]=None, # optional dict with  keys and functions called with filepath as argument. Results stored as metadata.
                         text_callables:Optional[dict]=None, # optional dict with  keys and functions called with file text as argument. Results stored as metadata.
                         **kwargs,
                         ) -> List[Document]:
    """
    Extract text from a single document. Will attempt to OCR PDFs, if necessary.


    Note that extra kwargs can be supplied to configure the behavior of PDF loaders.
    For instance, supplying `infer_table_structure` will cause `load_single_document` to try and
    infer and extract tables from PDFs. When `pdf_unstructured=True` and `infer_table_structure=True`,
    tables are represented as HTML within the main body of extracted text. In all other cases, inferred tables
    are represented as Markdown and appended to the end of the extracted text when `infer_table_structure=True`.
    """
    if pdf_unstructured and pdf_markdown:
        raise ValueError('pdf_unstructured and pdf_markdown cannot both be True.')
    file_callables = {} if not file_callables else file_callables
    text_callables = {} if not text_callables else text_callables
    
    # Normalize path for consistent handling - first get absolute path
    file_path = os.path.abspath(file_path)
    # Then normalize and standardize path separators to forward slashes for cross-platform consistency
    file_path = os.path.normpath(file_path).replace('\\', '/')


    # extract metadata
    file_metadata = {}
    if store_md5:
        file_metadata['md5'] = helpers.md5sum(file_path)
    if store_mimetype:
        file_metadata['mimetype'], _, _ = helpers.extract_mimetype(file_path)
    if store_file_dates:
        file_metadata['createdate'], file_metadata['modifydate'] = helpers.extract_file_dates(file_path)
    ext = helpers.extract_extension(file_path)
    file_metadata['extension'] = ext
    file_metadata.update(_apply_file_callables(file_path, file_callables))
        
    # load file
    if ext in LOADER_MAPPING:
        try:
            if ext == PDF:
                if pdf_unstructured:
                    ext = PDFOCR
                elif pdf_markdown:
                    ext = PDFMD
            loader_class, loader_args = LOADER_MAPPING[ext]
            loader_args = loader_args.copy() # copy so any supplied kwargs do not persist across calls
            if ext in PDF_EXTS:
                loader_args.update(kwargs)
            loader = loader_class(file_path, **loader_args)
            if ext in PDF_EXTS and ext != PDFOCR:
                with warnings.catch_warnings():
                    warnings.simplefilter(action='ignore', category=UserWarning)
                    docs = loader.load()
                    file_metadata.update(_apply_text_callables(docs, text_callables))
                    docs = _update_metadata(docs, file_metadata)
                if not docs or len('\n'.join([d.page_content.strip() for d in docs]).strip()) < OCR_CHAR_THRESH:
                    loader_class, loader_args = LOADER_MAPPING[PDFOCR]
                    loader = loader_class(file_path, **loader_args)
                    file_metadata['ocr'] = True
                    docs = loader.load()
                    for doc in docs:
                        doc.metadata['source'] = file_path
                    file_metadata.update(_apply_text_callables(docs, text_callables))
                    docs = _update_metadata(docs, file_metadata)
            else:
                docs = loader.load()
                file_metadata.update(_apply_text_callables(docs, text_callables))                
                docs = _update_metadata(docs, file_metadata)
            extra_keys = list(file_metadata.keys() | text_callables.keys())
            return helpers.set_metadata_defaults(docs, extra_keys=extra_keys)
        except Exception as e:
            logger.warning(f'\nSkipping {file_path} due to error: {str(e)}')
            import traceback
            print(traceback.format_exc())

    else:
        logger.warning(f"\nSkipping {file_path} due to unsupported file extension: '{ext}'")


def _ignore_file(file_path, ignored_files:List[str]=[], ignore_fn:Optional[Callable]=None):
    # Normalize path for consistent handling - first get absolute path, then normalize path separators
    file_path = os.path.abspath(file_path)
    file_path = os.path.normpath(file_path).replace('\\', '/')
    
    # Normalize ignored_files as well for consistent comparison
    normalized_ignored_files = [os.path.normpath(f).replace('\\', '/') for f in ignored_files]
    
    return file_path in normalized_ignored_files or \
            os.path.basename(file_path).startswith('~$') or \
            (ignore_fn is not None and ignore_fn(file_path))


def load_documents(source_dir: str, # path to folder containing documents
                   ignored_files: List[str] = [], # list of filepaths to ignore
                   ignore_fn:Optional[Callable] = None, # callable that accepts file path and returns True for ignored files
                   caption_tables:bool=False,# If True, agument table text with summaries of tables if infer_table_structure is True.
                   extract_document_titles:bool=False, # If True, infer document title and attach to individual chunks
                   llm=None, # a reference to the LLM (used by `caption_tables` and `extract_document_titles`
                   n_proc:Optional[int]=None, # number of CPU cores to use for text extraction. If None, use maximum for system.
                   verbose:bool=True, # verbosity
                   preserve_paragraphs:bool=False, # This is not used here and is only included to mask it from being forwarded to `load_single_document`.
                   **kwargs
) -> List[Document]:
    """
    Loads all documents from the source documents directory, ignoring specified files.
    Extra kwargs fed to `ingest.load_single_document`.

    Returns a generator over documents.
    """

    def keep(file_path):
        return not _ignore_file(file_path, ignored_files, ignore_fn)
    def all_files():
        for f in helpers.extract_files(source_dir, LOADER_MAPPING):
            yield f
    filtered_files = filtered_generator(all_files(), criteria=[keep])
    total = sum(1 for _ in filtered_generator(all_files(), criteria=[keep]))
    load_args = kwargs.copy()

    if n_proc==1:
        prog_msg = 'Loading documents...'
        mybar = None
        for i, filtered_file in enumerate(tqdm(filtered_files, total=total)):
            docs = load_single_document(filtered_file, **load_args)
            if docs is None: continue
            if llm and caption_tables:
                summarize_tables(docs, llm=llm, **kwargs)
            if llm and extract_document_titles:
                title = extract_title(docs, llm=llm, **kwargs)
                for doc in docs:
                    if title:
                        doc.metadata['document_title'] = title
            yield from docs
    else:
        if kwargs.get('infer_table_structure', False):
            # Use "spawn" if using TableTransformers
            # Reference: https://github.com/pytorch/pytorch/issues/40403
            multiprocessing.set_start_method('spawn', force=True)
            # call helpers.extract_tables sequentially below instead of in load_single_document if n_proc>1
            # because helpers.extract_tables is not well-suited to multiprocessing even with line above
            if not n_proc or n_proc>1:
                load_args = {k:load_args[k] for k in load_args if k!='infer_table_structure'}
        with multiprocessing.Pool(processes=n_proc if n_proc else os.cpu_count()) as pool:
            results = []
            with tqdm(
                total=total, desc="Loading new documents", ncols=80, disable=not verbose
            ) as pbar:
                for i, docs in enumerate(
                    pool.imap_unordered(functools.partial(load_single_document, **load_args),
                                                          filtered_files)
                ):
                    pbar.update()
                    if docs is None: continue
                    if kwargs.get('infer_table_structure', False):
                        docs = helpers.extract_tables(docs=docs)
                    if llm and caption_tables:
                        summarize_tables(docs, llm=llm, **kwargs)
                    if llm and extract_document_titles:
                        title = extract_title(docs, llm=llm, **kwargs)
                        for doc in docs:
                            if title:
                                doc.metadata['document_title'] = title
                    yield from docs
            # Make sure to close the pool when done
            pool.close()
            pool.join()


def process_folder(
    source_directory: str, # path to folder containing document store
    chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by `langchain.text_splitter.RecursiveCharacterTextSplitter`
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
    ignored_files: List[str] = [], # list of files to ignore
    ignore_fn:Optional[Callable] = None, # Callable that accepts the file path (including file name) as input and ignores if returns True
    batch_size:int=CHROMA_MAX, # batch size used when processing documents
    **kwargs


) -> List[Document]:
    """
    Load documents from folder, extract text from them, split texts into chunks.
    Extra kwargs fed to `ingest.load_documents` and `ingest.load_single_document`.
    """
    print(f"Loading documents from {source_directory}")
    documents = load_documents(source_directory,
                              ignored_files, ignore_fn=ignore_fn,
                              **kwargs)
    documents = list(documents)

    if not documents:
        print("No new documents to process")
        return

    batches = batch_list(documents, batch_size)
    total = sum(1 for _ in batch_list(documents, batch_size))
    for docs in tqdm(batches, total=total,
                     desc=f'Processing and chunking {len(documents)} new documents'):
        yield from chunk_documents(docs,
                                  chunk_size = chunk_size,
                                  chunk_overlap = chunk_overlap,
                                  **kwargs)


def chunk_documents(
    documents: list, # list of LangChain Documents
    chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by `langchain.text_splitter.RecursiveCharacterTextSplitter`
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
    infer_table_structure:bool = False, # This should be set to True if `documents` may contain contain tables (i.e., `doc.metadata['table']=True`).
    preserve_paragraphs:bool=False, # If True, strictly chunk by paragraph and only split if paragraph exceeds `chunk_size`. If False, small paragraphs will be accumulated into a single chunk until `chunk_size` is exceeded.
    **kwargs


) -> List[Document]:
    """
    Process list of Documents by splitting into chunks.
    """
    # remove tables before chunking
    if infer_table_structure and not kwargs.get('pdf_unstructured', False):
        tables = [d for d in documents if d.metadata.get('table', False)]
        docs = [d for d in documents if not d.metadata.get('table', False)]
    else:
        tables = []
        docs = documents

    # initialize the splitter
    contains_markdown = kwargs.get('pdf_markdown', False)
    if contains_markdown:
        text_splitter = RecursiveCharacterTextSplitter.from_language(
            language=Language.MARKDOWN,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap)
    elif preserve_paragraphs:
        text_splitter = helpers.ParagraphTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    else:
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

    # split non-table texts
    texts = text_splitter.split_documents(docs)

    # attempt to remove text chunks containing mangled tables
    texts = [d for d in texts if not helpers.includes_caption(d)]

    # split table texts
    if tables:
        table_splitter = RecursiveCharacterTextSplitter.from_language(
            language=Language.MARKDOWN,
            chunk_size=2000,
            chunk_overlap=0)
        table_texts = table_splitter.split_documents(tables)
        texts.extend(table_texts)

    # attach document title to each chunk (where title was extracted earlier by `load_documents`)
    if kwargs.get('extract_document_titles', False):
        for text in texts:
            if text.metadata.get('document_title', ''):
                text.page_content = f'The content below is from a document titled, \"{text.metadata["document_title"]}\"\n\n{text.page_content}'
    return texts


def does_vectorstore_exist(db) -> bool:
    """
    Checks if vectorstore exists
    """
    if not db.get()["documents"]:
        return False
    return True


def batchify_chunks(texts, batch_size=CHROMA_MAX):
    """
    split texts into batches specifically for Chroma
    """
    split_docs_chunked = batch_list(texts, batch_size)
    total_chunks = sum(1 for _ in batch_list(texts, batch_size))
    return split_docs_chunked, total_chunks





# %% ../../nbs/01_ingest.base.ipynb 6
from abc import ABC, abstractmethod

class VectorStore(ABC):

    def ingest(
        self,
        source_directory: str, # path to folder containing document store
        chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by [langchain.text_splitter.RecursiveCharacterTextSplitter](https://api.python.langchain.com/en/latest/character/langchain_text_splitters.character.RecursiveCharacterTextSplitter.html)
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
        ignore_fn:Optional[Callable] = None, # Optional function that accepts the file path (including file name) as input and returns `True` if file path should not be ingested.
        batch_size:int=CHROMA_MAX, # batch size used when processing documents
        **kwargs
    ) -> None:
        """
        Ingests all documents in `source_directory` (previously-ingested documents are
        ignored). When retrieved, the
        [Document](https://api.python.langchain.com/en/latest/documents/langchain_core.documents.base.Document.html)
        objects will each have a `metadata` dict with the absolute path to the file
        in `metadata["source"]`.
        Extra kwargs fed to `ingest.load_single_document`.
        """

        if not os.path.exists(source_directory):
            raise ValueError("The source_directory does not exist.")
        elif os.path.isfile(source_directory):
            raise ValueError(
                "The source_directory argument must be a folder, not a file."
            )
        texts = None
        if self.exists():
            # Update and store locally vectorstore
            print(f"Appending to existing vectorstore at {self.persist_directory}")
            # Get existing sources and normalize them for consistent comparison
            ignored_files = set([os.path.normpath(d['source']).replace('\\', '/') 
                                for d in self.get_all_docs()])
        else:
            print(f"Creating new vectorstore at {self.persist_directory}")
            ignored_files = []

        texts = process_folder(
            source_directory,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            ignored_files=ignored_files,
            ignore_fn=ignore_fn,
            batch_size=batch_size,
            **kwargs

        )

        texts = list(texts)
        print(f"Split into {len(texts)} chunks of text (max. {chunk_size} chars each for text; max. {TABLE_CHUNK_SIZE} chars for tables)")

        self.add_documents(texts, batch_size=batch_size)

        if texts:
            print(
                "Ingestion complete! You can now query your documents using the LLM.ask or LLM.chat methods"
            )
        db = None
        return
    
    def init_embedding_model(self, 
                             embedding_model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
                             embedding_model_kwargs: Optional[dict] = None,
                             embedding_encode_kwargs: dict = {"normalize_embeddings": False},
                             **kwargs
                             ):
        """
        Instantiate embedding model
        """
        if not embedding_model_kwargs:
            import torch
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            embedding_model_kwargs = {"device": device}          
        self.embeddings =  HuggingFaceEmbeddings(model_name=embedding_model_name, 
                                     model_kwargs=embedding_model_kwargs,
                                     encode_kwargs=embedding_encode_kwargs)

    def get_embedding_model(self):
        """
        Returns an instance to the `langchain_huggingface.HuggingFaceEmbeddings` instance
        """
        return self.embeddings


    def check(self):
        """
        Raise exception if `VectorStore.exists()` returns False
        """
        if not self.exists():
            raise Exception('The vector store is either empty or does not yet exist. '+\
                            'Please invoke the `ingest` method or `add_document` method.')

    @abstractmethod
    def get_db(self):
        """
        Get the raw, underlying vector database or search index.
        """
        pass


    @abstractmethod
    def exists(self):
        """
        Returns True if vector store has been initialized and contains documents.
        """
        pass


    @abstractmethod
    def add_documents(self, documents):
        """
        Stores instances of `langchain_core.documents.base.Document` in vector store
        """
        pass

    @abstractmethod
    def remove_document(self, id_to_delete):
        """
        Remove a single document.
        """
        pass
    
    @abstractmethod
    def update_documents(self, *args, **kwargs):
        """
        Update a set of documents.
        """
        pass

    @abstractmethod
    def get_all_docs(self):
        """
        Returns a list of files previously added to vector store.
        """
        pass

    @abstractmethod
    def get_doc(self, id):
        """
        Retrieve a document by ID
        """
        pass

    @abstractmethod
    def get_size(self):
        """
        Get total number of records added to vector store
        """
        pass

    @abstractmethod
    def erase(self):
        """
        Removes all documents in vector store
        """
        pass

    @abstractmethod
    def query(self, query):
        """
        Queries the vector store.
        For sparse stores, this is simply a keyword-search.
        For dense stores, this is equivalent to semantic_search.
        """
        pass

    @abstractmethod
    def semantic_search(self):
        """
        Semantic search of vector store
        """
        pass



"""Agent-based task execution pipeline that wraps smolagents to provide autonomous task completion capabilities using local LLMs."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/04_pipelines.agent.base.ipynb.

# %% auto 0
__all__ = ['Agent']

# %% ../../../nbs/04_pipelines.agent.base.ipynb 3
import onprem
from typing import Callable
import warnings
from ...ingest.base import VectorStore

from smolagents import PythonInterpreterTool, WebSearchTool, VisitWebpageTool, Tool as SA_Tool
from smolagents import ToolCallingAgent, CodeAgent
from .model import AgentModel
from . import tools as tool_utils

class Agent:
    """
    Pipeline for agent-based task execution using smolagents.
    Extra kwargs are supplied directly to agent instantation.
    
    Args:
        llm (LLM): An onprem LLM instance to use for agent reasoning
        agent_type (str, optional): Type of agent to use ('tool_calling' or 'code')
        max_steps (int, optional): Maximum number of steps the agent can take
        tools (dict, optional): a dictionary of tools for agent to use
    """
    

    
    def __init__(
        self, 
        llm: onprem.LLM,
        agent_type: str = "tool_calling",
        max_steps: int = 20,
        tools:dict = {},
        **kwargs,
    ):

        self.model = AgentModel(llm)
        self.tools = tools or {}
        self.agent_type = agent_type
        self.max_steps = max_steps
        self.kwargs = kwargs
                     
        # Initialize the agent
        self._init_agent()
    
    def _init_agent(self):
        """Initialize the appropriate smolagents agent type."""
        # Convert the LLM to a smolagents-compatible model
        
        # Get the tool list
        tool_list = list(self.tools.values())
        
        # Initialize the agent based on the agent type
        if self.agent_type == "tool_calling":
            self.agent = ToolCallingAgent(
                tools=tool_list,
                model=self.model,
                max_steps = self.max_steps,
                **self.kwargs
             )
        elif self.agent_type == "code":
            self.agent = CodeAgent(
                tools=tool_list,
                model=self.model,
                max_steps = self.max_steps,
                **self.kwargs
             )
        else:
            raise ValueError(f"Unsupported agent type: {self.agent_type}")
    
   
    def add_tool(self, name: str, tool_instance): # tool_instance is SA_Tool
        """
        Add a tool to the agent.
        
        Args:
            name (str): The name of the tool
            tool_instance (Tool): The tool instance
        """
        self.tools[name] = tool_instance
        self._init_agent()  # Reinitialize the agent with the new tool

    def add_default_tool(
        self, name:str
    ):
        """
        Create one of the built-in tools
                 
        Returns:
            Tool: The created tool
        """
        if name not in tool_utils.DEFAULT_TOOLS:
            raise ValueError(f'{name} is not one of the built-in tools: {tool_utils.DEFAULT_TOOLS.keys()}')
        self.add_tool(name, tool_utils.DEFAULT_TOOLS[name])
    
    def add_websearch_tool(
        self, 
    ):
        """
        Create a tool to perform Web searches.
                 
        Returns:
            Tool: The created tool
        """
        name = 'websearch'
        self.add_default_tool(name)

    
    def add_webview_tool(
        self, 
    ):
        """
        Create a tool to visit Web page
                 
        Returns:
            Tool: The created tool
        """
        name = 'webview'
        self.add_default_tool(name)

    
    def add_python_tool(
        self, 
    ):
        """
        Create a tool to access Python interpreter
                 
        Returns:
            Tool: The created tool
        """
        name = 'python'
        self.add_default_tool(name)

    
    def add_function_tool(
        self, 
        func: Callable, 
    ):
        """
        Create a tool from a function and its documentation.
        
        Args:
            func (Callable): The function to wrap as a tool
            name (str, optional): The name of the tool (defaults to function name)
            description (str, optional): The description of the tool (defaults to function docstring)
            
        Returns:
            Tool: The created tool
        """
        from types import FunctionType, MethodType

        if not isinstance(func, SA_Tool) and (isinstance(func, (FunctionType, MethodType)) or hasattr(func, "__call__")):
            name = func.__name__
            tool = tool_utils.createtool(func)
        else:
            raise ValueError(f'{func} is not a callable method or function')
        self.add_tool(name, tool)


    def add_vectorstore_tool(
        self, 
        name: str, 
        store: VectorStore, 
        description: str = "Search a vector database for relevant information",
    ):
        """
        Create a tool from a VectorStore instance.
        
        Args:
            name (str): The name of the vector store tool
            store (VectorStore): The vector store instance
            description (str, optional): The description of the vector store  
        Returns:
            Tool: The created tool
        """
        self.add_tool(name, tool_utils.VectorStoreTool(name, description, store))


    def add_mcp_tool(self, url:str):
        """
        Add tool to access MCP server
        """
        import mcpadapt.core
        from mcpadapt.smolagents_adapter import SmolAgentsAdapter
        
        mcp_tools = mcpadapt.core.MCPAdapt({"url": url}, SmolAgentsAdapter()).tools()
        for i, mcp_tool in enumerate(mcp_tools):
            self.add_tool(f'{url} (i)', mcp_tool)


    def run(self, task: str) -> str:
        """
        Run the agent on a given task.
        
        Args:
            task (str): The task description
            
        Returns:
            str: The agent's response
        """
        if not self.tools and self.agent_type == "tool_calling":
            raise ValueError('No tools have been added to agent. Please add at least one tool using on eof the Agent.add_* methods.')
            
        result = self.agent.run(task)
        return '\n'.join(result) if isinstance(result, list) else result



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/onprem',
                'doc_host': 'https://amaiya.github.io',
                'git_url': 'https://github.com/amaiya/onprem',
                'lib_path': 'onprem'},
  'syms': { 'onprem.app.OnPrem': {},
            'onprem.app.console': {},
            'onprem.app.pages.1_Prompts': {},
            'onprem.app.pages.2_Document_QA': {},
            'onprem.app.pages.3_Document_Analysis': {},
            'onprem.app.pages.4_Document_Search': {},
            'onprem.app.pages.5_Manage': {},
            'onprem.app.utils': {},
            'onprem.hf.base': { 'onprem.hf.base.Pipeline': ('hf.base.html#pipeline', 'onprem/hf/base.py'),
                                'onprem.hf.base.Pipeline.batch': ('hf.base.html#pipeline.batch', 'onprem/hf/base.py')},
            'onprem.hf.data.base': { 'onprem.hf.data.base.Data': ('hf.data.base.html#data', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.__call__': ('hf.data.base.html#data.__call__', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.__init__': ('hf.data.base.html#data.__init__', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.labels': ('hf.data.base.html#data.labels', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.length': ('hf.data.base.html#data.length', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.prepare': ('hf.data.base.html#data.prepare', 'onprem/hf/data/base.py'),
                                     'onprem.hf.data.base.Data.process': ('hf.data.base.html#data.process', 'onprem/hf/data/base.py')},
            'onprem.hf.data.labels': { 'onprem.hf.data.labels.Labels': ('hf.data.labels.html#labels', 'onprem/hf/data/labels.py'),
                                       'onprem.hf.data.labels.Labels.__init__': ( 'hf.data.labels.html#labels.__init__',
                                                                                  'onprem/hf/data/labels.py'),
                                       'onprem.hf.data.labels.Labels.process': ( 'hf.data.labels.html#labels.process',
                                                                                 'onprem/hf/data/labels.py')},
            'onprem.hf.data.questions': { 'onprem.hf.data.questions.Questions': ( 'hf.data.questions.html#questions',
                                                                                  'onprem/hf/data/questions.py'),
                                          'onprem.hf.data.questions.Questions.__init__': ( 'hf.data.questions.html#questions.__init__',
                                                                                           'onprem/hf/data/questions.py'),
                                          'onprem.hf.data.questions.Questions.answers': ( 'hf.data.questions.html#questions.answers',
                                                                                          'onprem/hf/data/questions.py'),
                                          'onprem.hf.data.questions.Questions.process': ( 'hf.data.questions.html#questions.process',
                                                                                          'onprem/hf/data/questions.py'),
                                          'onprem.hf.data.questions.Questions.tokenize': ( 'hf.data.questions.html#questions.tokenize',
                                                                                           'onprem/hf/data/questions.py')},
            'onprem.hf.data.sequences': { 'onprem.hf.data.sequences.Sequences': ( 'hf.data.sequences.html#sequences',
                                                                                  'onprem/hf/data/sequences.py'),
                                          'onprem.hf.data.sequences.Sequences.__init__': ( 'hf.data.sequences.html#sequences.__init__',
                                                                                           'onprem/hf/data/sequences.py'),
                                          'onprem.hf.data.sequences.Sequences.process': ( 'hf.data.sequences.html#sequences.process',
                                                                                          'onprem/hf/data/sequences.py')},
            'onprem.hf.data.texts': { 'onprem.hf.data.texts.Texts': ('hf.data.texts.html#texts', 'onprem/hf/data/texts.py'),
                                      'onprem.hf.data.texts.Texts.__init__': ( 'hf.data.texts.html#texts.__init__',
                                                                               'onprem/hf/data/texts.py'),
                                      'onprem.hf.data.texts.Texts.concat': ('hf.data.texts.html#texts.concat', 'onprem/hf/data/texts.py'),
                                      'onprem.hf.data.texts.Texts.process': ( 'hf.data.texts.html#texts.process',
                                                                              'onprem/hf/data/texts.py')},
            'onprem.hf.data.tokens': { 'onprem.hf.data.tokens.Tokens': ('hf.data.tokens.html#tokens', 'onprem/hf/data/tokens.py'),
                                       'onprem.hf.data.tokens.Tokens.__getitem__': ( 'hf.data.tokens.html#tokens.__getitem__',
                                                                                     'onprem/hf/data/tokens.py'),
                                       'onprem.hf.data.tokens.Tokens.__init__': ( 'hf.data.tokens.html#tokens.__init__',
                                                                                  'onprem/hf/data/tokens.py'),
                                       'onprem.hf.data.tokens.Tokens.__len__': ( 'hf.data.tokens.html#tokens.__len__',
                                                                                 'onprem/hf/data/tokens.py')},
            'onprem.hf.models.models': { 'onprem.hf.models.models.Models': ('hf.models.models.html#models', 'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.checklength': ( 'hf.models.models.html#models.checklength',
                                                                                         'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.device': ( 'hf.models.models.html#models.device',
                                                                                    'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.deviceid': ( 'hf.models.models.html#models.deviceid',
                                                                                      'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.finddevice': ( 'hf.models.models.html#models.finddevice',
                                                                                        'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.hasaccelerator': ( 'hf.models.models.html#models.hasaccelerator',
                                                                                            'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.hasmpsdevice': ( 'hf.models.models.html#models.hasmpsdevice',
                                                                                          'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.load': ( 'hf.models.models.html#models.load',
                                                                                  'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.maxlength': ( 'hf.models.models.html#models.maxlength',
                                                                                       'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.reference': ( 'hf.models.models.html#models.reference',
                                                                                       'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.task': ( 'hf.models.models.html#models.task',
                                                                                  'onprem/hf/models/models.py'),
                                         'onprem.hf.models.models.Models.tokenizer': ( 'hf.models.models.html#models.tokenizer',
                                                                                       'onprem/hf/models/models.py')},
            'onprem.hf.models.onnx': { 'onprem.hf.models.onnx.OnnxConfig': ('hf.models.onnx.html#onnxconfig', 'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel': ('hf.models.onnx.html#onnxmodel', 'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel.__init__': ( 'hf.models.onnx.html#onnxmodel.__init__',
                                                                                     'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel.device': ( 'hf.models.onnx.html#onnxmodel.device',
                                                                                   'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel.forward': ( 'hf.models.onnx.html#onnxmodel.forward',
                                                                                    'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel.parse': ( 'hf.models.onnx.html#onnxmodel.parse',
                                                                                  'onprem/hf/models/onnx.py'),
                                       'onprem.hf.models.onnx.OnnxModel.providers': ( 'hf.models.onnx.html#onnxmodel.providers',
                                                                                      'onprem/hf/models/onnx.py')},
            'onprem.hf.models.pooling.base': { 'onprem.hf.models.pooling.base.Pooling': ( 'hf.models.pooling.base.html#pooling',
                                                                                          'onprem/hf/models/pooling/base.py'),
                                               'onprem.hf.models.pooling.base.Pooling.__init__': ( 'hf.models.pooling.base.html#pooling.__init__',
                                                                                                   'onprem/hf/models/pooling/base.py'),
                                               'onprem.hf.models.pooling.base.Pooling.chunk': ( 'hf.models.pooling.base.html#pooling.chunk',
                                                                                                'onprem/hf/models/pooling/base.py'),
                                               'onprem.hf.models.pooling.base.Pooling.encode': ( 'hf.models.pooling.base.html#pooling.encode',
                                                                                                 'onprem/hf/models/pooling/base.py'),
                                               'onprem.hf.models.pooling.base.Pooling.forward': ( 'hf.models.pooling.base.html#pooling.forward',
                                                                                                  'onprem/hf/models/pooling/base.py')},
            'onprem.hf.models.pooling.cls': { 'onprem.hf.models.pooling.cls.ClsPooling': ( 'hf.models.pooling.cls.html#clspooling',
                                                                                           'onprem/hf/models/pooling/cls.py'),
                                              'onprem.hf.models.pooling.cls.ClsPooling.forward': ( 'hf.models.pooling.cls.html#clspooling.forward',
                                                                                                   'onprem/hf/models/pooling/cls.py')},
            'onprem.hf.models.pooling.factory': { 'onprem.hf.models.pooling.factory.PoolingFactory': ( 'hf.models.pooling.factory.html#poolingfactory',
                                                                                                       'onprem/hf/models/pooling/factory.py'),
                                                  'onprem.hf.models.pooling.factory.PoolingFactory.create': ( 'hf.models.pooling.factory.html#poolingfactory.create',
                                                                                                              'onprem/hf/models/pooling/factory.py'),
                                                  'onprem.hf.models.pooling.factory.PoolingFactory.load': ( 'hf.models.pooling.factory.html#poolingfactory.load',
                                                                                                            'onprem/hf/models/pooling/factory.py'),
                                                  'onprem.hf.models.pooling.factory.PoolingFactory.maxlength': ( 'hf.models.pooling.factory.html#poolingfactory.maxlength',
                                                                                                                 'onprem/hf/models/pooling/factory.py'),
                                                  'onprem.hf.models.pooling.factory.PoolingFactory.method': ( 'hf.models.pooling.factory.html#poolingfactory.method',
                                                                                                              'onprem/hf/models/pooling/factory.py')},
            'onprem.hf.models.pooling.mean': { 'onprem.hf.models.pooling.mean.MeanPooling': ( 'hf.models.pooling.mean.html#meanpooling',
                                                                                              'onprem/hf/models/pooling/mean.py'),
                                               'onprem.hf.models.pooling.mean.MeanPooling.forward': ( 'hf.models.pooling.mean.html#meanpooling.forward',
                                                                                                      'onprem/hf/models/pooling/mean.py')},
            'onprem.hf.models.registry': { 'onprem.hf.models.registry.Registry': ( 'hf.models.registry.html#registry',
                                                                                   'onprem/hf/models/registry.py'),
                                           'onprem.hf.models.registry.Registry.register': ( 'hf.models.registry.html#registry.register',
                                                                                            'onprem/hf/models/registry.py')},
            'onprem.hf.models.tokendetection': { 'onprem.hf.models.tokendetection.TokenDetection': ( 'hf.models.tokendetection.html#tokendetection',
                                                                                                     'onprem/hf/models/tokendetection.py'),
                                                 'onprem.hf.models.tokendetection.TokenDetection.__init__': ( 'hf.models.tokendetection.html#tokendetection.__init__',
                                                                                                              'onprem/hf/models/tokendetection.py'),
                                                 'onprem.hf.models.tokendetection.TokenDetection.forward': ( 'hf.models.tokendetection.html#tokendetection.forward',
                                                                                                             'onprem/hf/models/tokendetection.py'),
                                                 'onprem.hf.models.tokendetection.TokenDetection.save_pretrained': ( 'hf.models.tokendetection.html#tokendetection.save_pretrained',
                                                                                                                     'onprem/hf/models/tokendetection.py')},
            'onprem.hf.tensors': { 'onprem.hf.tensors.Tensors': ('hf.tensors.html#tensors', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.context': ('hf.tensors.html#tensors.context', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.quantize': ('hf.tensors.html#tensors.quantize', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.tensor': ('hf.tensors.html#tensors.tensor', 'onprem/hf/tensors.py')},
            'onprem.hf.train.hfonnx': { 'onprem.hf.train.hfonnx.HFOnnx': ('hf.train.hfonnx.html#hfonnx', 'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.HFOnnx.__call__': ( 'hf.train.hfonnx.html#hfonnx.__call__',
                                                                                    'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.HFOnnx.parameters': ( 'hf.train.hfonnx.html#hfonnx.parameters',
                                                                                      'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.HFOnnx.quantization': ( 'hf.train.hfonnx.html#hfonnx.quantization',
                                                                                        'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.PoolingOnnx': ( 'hf.train.hfonnx.html#poolingonnx',
                                                                                'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.PoolingOnnx.__init__': ( 'hf.train.hfonnx.html#poolingonnx.__init__',
                                                                                         'onprem/hf/train/hfonnx.py'),
                                        'onprem.hf.train.hfonnx.PoolingOnnx.forward': ( 'hf.train.hfonnx.html#poolingonnx.forward',
                                                                                        'onprem/hf/train/hfonnx.py')},
            'onprem.hf.train.hftrainer': { 'onprem.hf.train.hftrainer.HFTrainer': ( 'hf.train.hftrainer.html#hftrainer',
                                                                                    'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.__call__': ( 'hf.train.hftrainer.html#hftrainer.__call__',
                                                                                             'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.load': ( 'hf.train.hftrainer.html#hftrainer.load',
                                                                                         'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.lora': ( 'hf.train.hftrainer.html#hftrainer.lora',
                                                                                         'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.loratask': ( 'hf.train.hftrainer.html#hftrainer.loratask',
                                                                                             'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.model': ( 'hf.train.hftrainer.html#hftrainer.model',
                                                                                          'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.parse': ( 'hf.train.hftrainer.html#hftrainer.parse',
                                                                                          'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.peft': ( 'hf.train.hftrainer.html#hftrainer.peft',
                                                                                         'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.prepare': ( 'hf.train.hftrainer.html#hftrainer.prepare',
                                                                                            'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.HFTrainer.quantization': ( 'hf.train.hftrainer.html#hftrainer.quantization',
                                                                                                 'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.TrainingArguments': ( 'hf.train.hftrainer.html#trainingarguments',
                                                                                            'onprem/hf/train/hftrainer.py'),
                                           'onprem.hf.train.hftrainer.TrainingArguments.should_save': ( 'hf.train.hftrainer.html#trainingarguments.should_save',
                                                                                                        'onprem/hf/train/hftrainer.py')},
            'onprem.hf.train.mlonnx': { 'onprem.hf.train.mlonnx.MLOnnx': ('hf.train.mlonnx.html#mlonnx', 'onprem/hf/train/mlonnx.py'),
                                        'onprem.hf.train.mlonnx.MLOnnx.__call__': ( 'hf.train.mlonnx.html#mlonnx.__call__',
                                                                                    'onprem/hf/train/mlonnx.py'),
                                        'onprem.hf.train.mlonnx.MLOnnx.__init__': ( 'hf.train.mlonnx.html#mlonnx.__init__',
                                                                                    'onprem/hf/train/mlonnx.py')},
            'onprem.ingest.base': { 'onprem.ingest.base.MyElmLoader': ('ingest.base.html#myelmloader', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.MyElmLoader.load': ('ingest.base.html#myelmloader.load', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.MyUnstructuredPDFLoader': ( 'ingest.base.html#myunstructuredpdfloader',
                                                                                    'onprem/ingest/base.py'),
                                    'onprem.ingest.base.MyUnstructuredPDFLoader.load': ( 'ingest.base.html#myunstructuredpdfloader.load',
                                                                                         'onprem/ingest/base.py'),
                                    'onprem.ingest.base.PDF2MarkdownLoader': ( 'ingest.base.html#pdf2markdownloader',
                                                                               'onprem/ingest/base.py'),
                                    'onprem.ingest.base.PDF2MarkdownLoader.load': ( 'ingest.base.html#pdf2markdownloader.load',
                                                                                    'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore': ('ingest.base.html#vectorstore', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.add_documents': ( 'ingest.base.html#vectorstore.add_documents',
                                                                                      'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.check': ('ingest.base.html#vectorstore.check', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.erase': ('ingest.base.html#vectorstore.erase', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.exists': ( 'ingest.base.html#vectorstore.exists',
                                                                               'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.get_all_docs': ( 'ingest.base.html#vectorstore.get_all_docs',
                                                                                     'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.get_db': ( 'ingest.base.html#vectorstore.get_db',
                                                                               'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.get_doc': ( 'ingest.base.html#vectorstore.get_doc',
                                                                                'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.get_embedding_model': ( 'ingest.base.html#vectorstore.get_embedding_model',
                                                                                            'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.get_size': ( 'ingest.base.html#vectorstore.get_size',
                                                                                 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.ingest': ( 'ingest.base.html#vectorstore.ingest',
                                                                               'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.init_embedding_model': ( 'ingest.base.html#vectorstore.init_embedding_model',
                                                                                             'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.query': ('ingest.base.html#vectorstore.query', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.remove_document': ( 'ingest.base.html#vectorstore.remove_document',
                                                                                        'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.semantic_search': ( 'ingest.base.html#vectorstore.semantic_search',
                                                                                        'onprem/ingest/base.py'),
                                    'onprem.ingest.base.VectorStore.update_documents': ( 'ingest.base.html#vectorstore.update_documents',
                                                                                         'onprem/ingest/base.py'),
                                    'onprem.ingest.base._PyMuPDFLoader': ('ingest.base.html#_pymupdfloader', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base._PyMuPDFLoader.load': ( 'ingest.base.html#_pymupdfloader.load',
                                                                                'onprem/ingest/base.py'),
                                    'onprem.ingest.base._apply_file_callables': ( 'ingest.base.html#_apply_file_callables',
                                                                                  'onprem/ingest/base.py'),
                                    'onprem.ingest.base._apply_text_callables': ( 'ingest.base.html#_apply_text_callables',
                                                                                  'onprem/ingest/base.py'),
                                    'onprem.ingest.base._ignore_file': ('ingest.base.html#_ignore_file', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base._update_metadata': ('ingest.base.html#_update_metadata', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.batchify_chunks': ('ingest.base.html#batchify_chunks', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.chunk_documents': ('ingest.base.html#chunk_documents', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.does_vectorstore_exist': ( 'ingest.base.html#does_vectorstore_exist',
                                                                                   'onprem/ingest/base.py'),
                                    'onprem.ingest.base.load_documents': ('ingest.base.html#load_documents', 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.load_single_document': ( 'ingest.base.html#load_single_document',
                                                                                 'onprem/ingest/base.py'),
                                    'onprem.ingest.base.process_folder': ('ingest.base.html#process_folder', 'onprem/ingest/base.py')},
            'onprem.ingest.helpers': { 'onprem.ingest.helpers.ParagraphTextSplitter': ( 'ingest.helpers.html#paragraphtextsplitter',
                                                                                        'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.ParagraphTextSplitter.__init__': ( 'ingest.helpers.html#paragraphtextsplitter.__init__',
                                                                                                 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.ParagraphTextSplitter._split_large_paragraph': ( 'ingest.helpers.html#paragraphtextsplitter._split_large_paragraph',
                                                                                                               'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.ParagraphTextSplitter.split_text': ( 'ingest.helpers.html#paragraphtextsplitter.split_text',
                                                                                                   'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers._apply_file_callables': ( 'ingest.helpers.html#_apply_file_callables',
                                                                                        'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.clean_text': ('ingest.helpers.html#clean_text', 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.create_document': ( 'ingest.helpers.html#create_document',
                                                                                  'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.date2iso': ('ingest.helpers.html#date2iso', 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.doc_from_dict': ( 'ingest.helpers.html#doc_from_dict',
                                                                                'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_extension': ( 'ingest.helpers.html#extract_extension',
                                                                                    'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_file_dates': ( 'ingest.helpers.html#extract_file_dates',
                                                                                     'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_file_metadata': ( 'ingest.helpers.html#extract_file_metadata',
                                                                                        'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_files': ( 'ingest.helpers.html#extract_files',
                                                                                'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_mimetype': ( 'ingest.helpers.html#extract_mimetype',
                                                                                   'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.extract_tables': ( 'ingest.helpers.html#extract_tables',
                                                                                 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.get_mimetype': ( 'ingest.helpers.html#get_mimetype',
                                                                               'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.includes_caption': ( 'ingest.helpers.html#includes_caption',
                                                                                   'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.is_random_plaintext': ( 'ingest.helpers.html#is_random_plaintext',
                                                                                      'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.iso2date': ('ingest.helpers.html#iso2date', 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.md5sum': ('ingest.helpers.html#md5sum', 'onprem/ingest/helpers.py'),
                                       'onprem.ingest.helpers.set_metadata_defaults': ( 'ingest.helpers.html#set_metadata_defaults',
                                                                                        'onprem/ingest/helpers.py')},
            'onprem.ingest.pdftables': { 'onprem.ingest.pdftables.PDFTables': ( 'ingest.pdftables.html#pdftables',
                                                                                'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.__init__': ( 'ingest.pdftables.html#pdftables.__init__',
                                                                                         'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.from_file': ( 'ingest.pdftables.html#pdftables.from_file',
                                                                                          'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.get_captions': ( 'ingest.pdftables.html#pdftables.get_captions',
                                                                                             'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.get_markdown_tables': ( 'ingest.pdftables.html#pdftables.get_markdown_tables',
                                                                                                    'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.get_tables': ( 'ingest.pdftables.html#pdftables.get_tables',
                                                                                           'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.PDFTables.to_markdown': ( 'ingest.pdftables.html#pdftables.to_markdown',
                                                                                            'onprem/ingest/pdftables.py'),
                                         'onprem.ingest.pdftables.ingest_pdf': ( 'ingest.pdftables.html#ingest_pdf',
                                                                                 'onprem/ingest/pdftables.py')},
            'onprem.ingest.stores.dense': { 'onprem.ingest.stores.dense.DenseStore': ( 'ingest.stores.dense.html#densestore',
                                                                                       'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.__init__': ( 'ingest.stores.dense.html#densestore.__init__',
                                                                                                'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore._convert_to_dict': ( 'ingest.stores.dense.html#densestore._convert_to_dict',
                                                                                                        'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.add_documents': ( 'ingest.stores.dense.html#densestore.add_documents',
                                                                                                     'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.erase': ( 'ingest.stores.dense.html#densestore.erase',
                                                                                             'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.exists': ( 'ingest.stores.dense.html#densestore.exists',
                                                                                              'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.get_all_docs': ( 'ingest.stores.dense.html#densestore.get_all_docs',
                                                                                                    'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.get_db': ( 'ingest.stores.dense.html#densestore.get_db',
                                                                                              'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.get_doc': ( 'ingest.stores.dense.html#densestore.get_doc',
                                                                                               'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.get_size': ( 'ingest.stores.dense.html#densestore.get_size',
                                                                                                'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.optimize_for_search': ( 'ingest.stores.dense.html#densestore.optimize_for_search',
                                                                                                           'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.query': ( 'ingest.stores.dense.html#densestore.query',
                                                                                             'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.remove_document': ( 'ingest.stores.dense.html#densestore.remove_document',
                                                                                                       'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.remove_source': ( 'ingest.stores.dense.html#densestore.remove_source',
                                                                                                     'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.semantic_search': ( 'ingest.stores.dense.html#densestore.semantic_search',
                                                                                                       'onprem/ingest/stores/dense.py'),
                                            'onprem.ingest.stores.dense.DenseStore.update_documents': ( 'ingest.stores.dense.html#densestore.update_documents',
                                                                                                        'onprem/ingest/stores/dense.py')},
            'onprem.ingest.stores.dual': { 'onprem.ingest.stores.dual.DualStore': ( 'ingest.stores.dual.html#dualstore',
                                                                                    'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.__init__': ( 'ingest.stores.dual.html#dualstore.__init__',
                                                                                             'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.add_documents': ( 'ingest.stores.dual.html#dualstore.add_documents',
                                                                                                  'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.erase': ( 'ingest.stores.dual.html#dualstore.erase',
                                                                                          'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.exists': ( 'ingest.stores.dual.html#dualstore.exists',
                                                                                           'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_all_docs': ( 'ingest.stores.dual.html#dualstore.get_all_docs',
                                                                                                 'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_db': ( 'ingest.stores.dual.html#dualstore.get_db',
                                                                                           'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_dense_db': ( 'ingest.stores.dual.html#dualstore.get_dense_db',
                                                                                                 'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_doc': ( 'ingest.stores.dual.html#dualstore.get_doc',
                                                                                            'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_size': ( 'ingest.stores.dual.html#dualstore.get_size',
                                                                                             'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.get_sparse_db': ( 'ingest.stores.dual.html#dualstore.get_sparse_db',
                                                                                                  'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.keyword_search': ( 'ingest.stores.dual.html#dualstore.keyword_search',
                                                                                                   'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.optimize_for_search': ( 'ingest.stores.dual.html#dualstore.optimize_for_search',
                                                                                                        'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.query': ( 'ingest.stores.dual.html#dualstore.query',
                                                                                          'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.remove_document': ( 'ingest.stores.dual.html#dualstore.remove_document',
                                                                                                    'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.remove_source': ( 'ingest.stores.dual.html#dualstore.remove_source',
                                                                                                  'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.semantic_search': ( 'ingest.stores.dual.html#dualstore.semantic_search',
                                                                                                    'onprem/ingest/stores/dual.py'),
                                           'onprem.ingest.stores.dual.DualStore.update_documents': ( 'ingest.stores.dual.html#dualstore.update_documents',
                                                                                                     'onprem/ingest/stores/dual.py')},
            'onprem.ingest.stores.sparse': { 'onprem.ingest.stores.sparse.SparseStore': ( 'ingest.stores.sparse.html#sparsestore',
                                                                                          'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.__init__': ( 'ingest.stores.sparse.html#sparsestore.__init__',
                                                                                                   'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore._preprocess_query': ( 'ingest.stores.sparse.html#sparsestore._preprocess_query',
                                                                                                            'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.add_documents': ( 'ingest.stores.sparse.html#sparsestore.add_documents',
                                                                                                        'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.delete_by_prefix': ( 'ingest.stores.sparse.html#sparsestore.delete_by_prefix',
                                                                                                           'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.doc2dict': ( 'ingest.stores.sparse.html#sparsestore.doc2dict',
                                                                                                   'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.erase': ( 'ingest.stores.sparse.html#sparsestore.erase',
                                                                                                'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.exists': ( 'ingest.stores.sparse.html#sparsestore.exists',
                                                                                                 'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.get_all_docs': ( 'ingest.stores.sparse.html#sparsestore.get_all_docs',
                                                                                                       'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.get_db': ( 'ingest.stores.sparse.html#sparsestore.get_db',
                                                                                                 'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.get_doc': ( 'ingest.stores.sparse.html#sparsestore.get_doc',
                                                                                                  'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.get_size': ( 'ingest.stores.sparse.html#sparsestore.get_size',
                                                                                                   'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.index_exists_in': ( 'ingest.stores.sparse.html#sparsestore.index_exists_in',
                                                                                                          'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.initialize_index': ( 'ingest.stores.sparse.html#sparsestore.initialize_index',
                                                                                                           'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.normalize_text': ( 'ingest.stores.sparse.html#sparsestore.normalize_text',
                                                                                                         'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.query': ( 'ingest.stores.sparse.html#sparsestore.query',
                                                                                                'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.remove_document': ( 'ingest.stores.sparse.html#sparsestore.remove_document',
                                                                                                          'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.remove_source': ( 'ingest.stores.sparse.html#sparsestore.remove_source',
                                                                                                        'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.semantic_search': ( 'ingest.stores.sparse.html#sparsestore.semantic_search',
                                                                                                          'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.SparseStore.update_documents': ( 'ingest.stores.sparse.html#sparsestore.update_documents',
                                                                                                           'onprem/ingest/stores/sparse.py'),
                                             'onprem.ingest.stores.sparse.default_schema': ( 'ingest.stores.sparse.html#default_schema',
                                                                                             'onprem/ingest/stores/sparse.py')},
            'onprem.llm.base': { 'onprem.llm.base.LLM': ('llm.base.html#llm', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.__init__': ('llm.base.html#llm.__init__', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM._ask': ('llm.base.html#llm._ask', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM._format_image_prompt': ( 'llm.base.html#llm._format_image_prompt',
                                                                               'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM._format_pydantic_prompt': ( 'llm.base.html#llm._format_pydantic_prompt',
                                                                                  'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.ask': ('llm.base.html#llm.ask', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.chat': ('llm.base.html#llm.chat', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.check_model': ('llm.base.html#llm.check_model', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.download_model': ('llm.base.html#llm.download_model', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.get_store_type': ('llm.base.html#llm.get_store_type', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.ingest': ('llm.base.html#llm.ingest', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_azure': ('llm.base.html#llm.is_azure', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_dense_store': ('llm.base.html#llm.is_dense_store', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_dual_store': ('llm.base.html#llm.is_dual_store', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_hf': ('llm.base.html#llm.is_hf', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_llamacpp': ('llm.base.html#llm.is_llamacpp', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_local': ('llm.base.html#llm.is_local', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_local_api': ('llm.base.html#llm.is_local_api', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_openai_model': ('llm.base.html#llm.is_openai_model', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.is_sparse_store': ('llm.base.html#llm.is_sparse_store', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.load_chatbot': ('llm.base.html#llm.load_chatbot', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.load_llm': ('llm.base.html#llm.load_llm', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.load_vectordb': ('llm.base.html#llm.load_vectordb', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.load_vectorstore': ('llm.base.html#llm.load_vectorstore', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.optimize_vectorstore_for_search': ( 'llm.base.html#llm.optimize_vectorstore_for_search',
                                                                                          'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.process_service': ('llm.base.html#llm.process_service', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.prompt': ('llm.base.html#llm.prompt', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.pydantic_prompt': ('llm.base.html#llm.pydantic_prompt', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.query': ('llm.base.html#llm.query', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.set_store_type': ('llm.base.html#llm.set_store_type', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.update_max_tokens': ('llm.base.html#llm.update_max_tokens', 'onprem/llm/base.py'),
                                 'onprem.llm.base.LLM.update_stop': ('llm.base.html#llm.update_stop', 'onprem/llm/base.py')},
            'onprem.llm.helpers': { 'onprem.llm.helpers.TableSummary': ('llm.helpers.html#tablesummary', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.Title': ('llm.helpers.html#title', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers._marshal_llm_to_json': ( 'llm.helpers.html#_marshal_llm_to_json',
                                                                                 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.caption_table_text': ( 'llm.helpers.html#caption_table_text',
                                                                               'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.decompose_question': ( 'llm.helpers.html#decompose_question',
                                                                               'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.extract_json': ('llm.helpers.html#extract_json', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.extract_title': ('llm.helpers.html#extract_title', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.needs_followup': ('llm.helpers.html#needs_followup', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.parse_code_markdown': ( 'llm.helpers.html#parse_code_markdown',
                                                                                'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.parse_json_markdown': ( 'llm.helpers.html#parse_json_markdown',
                                                                                'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.summarize_tables': ('llm.helpers.html#summarize_tables', 'onprem/llm/helpers.py'),
                                    'onprem.llm.helpers.truncate_prompt': ('llm.helpers.html#truncate_prompt', 'onprem/llm/helpers.py')},
            'onprem.pipelines.agent.base': { 'onprem.pipelines.agent.base.Agent': ( 'pipelines.agent.base.html#agent',
                                                                                    'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.__init__': ( 'pipelines.agent.base.html#agent.__init__',
                                                                                             'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent._init_agent': ( 'pipelines.agent.base.html#agent._init_agent',
                                                                                                'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_default_tool': ( 'pipelines.agent.base.html#agent.add_default_tool',
                                                                                                     'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_function_tool': ( 'pipelines.agent.base.html#agent.add_function_tool',
                                                                                                      'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_mcp_tool': ( 'pipelines.agent.base.html#agent.add_mcp_tool',
                                                                                                 'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_python_tool': ( 'pipelines.agent.base.html#agent.add_python_tool',
                                                                                                    'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_tool': ( 'pipelines.agent.base.html#agent.add_tool',
                                                                                             'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_vectorstore_tool': ( 'pipelines.agent.base.html#agent.add_vectorstore_tool',
                                                                                                         'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_websearch_tool': ( 'pipelines.agent.base.html#agent.add_websearch_tool',
                                                                                                       'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.add_webview_tool': ( 'pipelines.agent.base.html#agent.add_webview_tool',
                                                                                                     'onprem/pipelines/agent/base.py'),
                                             'onprem.pipelines.agent.base.Agent.run': ( 'pipelines.agent.base.html#agent.run',
                                                                                        'onprem/pipelines/agent/base.py')},
            'onprem.pipelines.agent.model': { 'onprem.pipelines.agent.model.AgentModel': ( 'pipelines.agent.model.html#agentmodel',
                                                                                           'onprem/pipelines/agent/model.py'),
                                              'onprem.pipelines.agent.model.AgentModel.__init__': ( 'pipelines.agent.model.html#agentmodel.__init__',
                                                                                                    'onprem/pipelines/agent/model.py'),
                                              'onprem.pipelines.agent.model.AgentModel.clean': ( 'pipelines.agent.model.html#agentmodel.clean',
                                                                                                 'onprem/pipelines/agent/model.py'),
                                              'onprem.pipelines.agent.model.AgentModel.generate': ( 'pipelines.agent.model.html#agentmodel.generate',
                                                                                                    'onprem/pipelines/agent/model.py'),
                                              'onprem.pipelines.agent.model.AgentModel.parameters': ( 'pipelines.agent.model.html#agentmodel.parameters',
                                                                                                      'onprem/pipelines/agent/model.py')},
            'onprem.pipelines.agent.tools': { 'onprem.pipelines.agent.tools.FunctionTool': ( 'pipelines.agent.tools.html#functiontool',
                                                                                             'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.FunctionTool.__init__': ( 'pipelines.agent.tools.html#functiontool.__init__',
                                                                                                      'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.FunctionTool.forward': ( 'pipelines.agent.tools.html#functiontool.forward',
                                                                                                     'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.VectorStoreTool': ( 'pipelines.agent.tools.html#vectorstoretool',
                                                                                                'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.VectorStoreTool.__init__': ( 'pipelines.agent.tools.html#vectorstoretool.__init__',
                                                                                                         'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.VectorStoreTool.forward': ( 'pipelines.agent.tools.html#vectorstoretool.forward',
                                                                                                        'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.createtool': ( 'pipelines.agent.tools.html#createtool',
                                                                                           'onprem/pipelines/agent/tools.py'),
                                              'onprem.pipelines.agent.tools.fromdocs': ( 'pipelines.agent.tools.html#fromdocs',
                                                                                         'onprem/pipelines/agent/tools.py')},
            'onprem.pipelines.classifier': { 'onprem.pipelines.classifier.ClassifierBase': ( 'pipelines.classifier.html#classifierbase',
                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.arrays2dataset': ( 'pipelines.classifier.html#classifierbase.arrays2dataset',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.dataset2arrays': ( 'pipelines.classifier.html#classifierbase.dataset2arrays',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.evaluate': ( 'pipelines.classifier.html#classifierbase.evaluate',
                                                                                                      'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.explain': ( 'pipelines.classifier.html#classifierbase.explain',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_explain_predictor': ( 'pipelines.classifier.html#classifierbase.get_explain_predictor',
                                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_labels': ( 'pipelines.classifier.html#classifierbase.get_labels',
                                                                                                        'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_trainer': ( 'pipelines.classifier.html#classifierbase.get_trainer',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.sample_examples': ( 'pipelines.classifier.html#classifierbase.sample_examples',
                                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.save': ( 'pipelines.classifier.html#classifierbase.save',
                                                                                                  'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.train': ( 'pipelines.classifier.html#classifierbase.train',
                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier': ( 'pipelines.classifier.html#fewshotclassifier',
                                                                                                'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.__init__': ( 'pipelines.classifier.html#fewshotclassifier.__init__',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier._get_explain_predictor': ( 'pipelines.classifier.html#fewshotclassifier._get_explain_predictor',
                                                                                                                       'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.predict': ( 'pipelines.classifier.html#fewshotclassifier.predict',
                                                                                                        'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.predict_proba': ( 'pipelines.classifier.html#fewshotclassifier.predict_proba',
                                                                                                              'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.save': ( 'pipelines.classifier.html#fewshotclassifier.save',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.train': ( 'pipelines.classifier.html#fewshotclassifier.train',
                                                                                                      'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier': ( 'pipelines.classifier.html#hfclassifier',
                                                                                           'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier.__init__': ( 'pipelines.classifier.html#hfclassifier.__init__',
                                                                                                    'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier._get_explain_predictor': ( 'pipelines.classifier.html#hfclassifier._get_explain_predictor',
                                                                                                                  'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier._get_pipeline': ( 'pipelines.classifier.html#hfclassifier._get_pipeline',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier._predict': ( 'pipelines.classifier.html#hfclassifier._predict',
                                                                                                    'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier._tempsave': ( 'pipelines.classifier.html#hfclassifier._tempsave',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier.predict': ( 'pipelines.classifier.html#hfclassifier.predict',
                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier.predict_proba': ( 'pipelines.classifier.html#hfclassifier.predict_proba',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier.save': ( 'pipelines.classifier.html#hfclassifier.save',
                                                                                                'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.HFClassifier.train': ( 'pipelines.classifier.html#hfclassifier.train',
                                                                                                 'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier': ( 'pipelines.classifier.html#skclassifier',
                                                                                           'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier.__init__': ( 'pipelines.classifier.html#skclassifier.__init__',
                                                                                                    'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier.predict': ( 'pipelines.classifier.html#skclassifier.predict',
                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier.predict_proba': ( 'pipelines.classifier.html#skclassifier.predict_proba',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier.save': ( 'pipelines.classifier.html#skclassifier.save',
                                                                                                'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.SKClassifier.train': ( 'pipelines.classifier.html#skclassifier.train',
                                                                                                 'onprem/pipelines/classifier.py')},
            'onprem.pipelines.extractor': { 'onprem.pipelines.extractor.Extractor': ( 'pipelines.extractor.html#extractor',
                                                                                      'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.__init__': ( 'pipelines.extractor.html#extractor.__init__',
                                                                                               'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.apply': ( 'pipelines.extractor.html#extractor.apply',
                                                                                            'onprem/pipelines/extractor.py')},
            'onprem.pipelines.guider': { 'onprem.pipelines.guider.Guider': ('pipelines.guider.html#guider', 'onprem/pipelines/guider.py'),
                                         'onprem.pipelines.guider.Guider.__init__': ( 'pipelines.guider.html#guider.__init__',
                                                                                      'onprem/pipelines/guider.py'),
                                         'onprem.pipelines.guider.Guider.prompt': ( 'pipelines.guider.html#guider.prompt',
                                                                                    'onprem/pipelines/guider.py')},
            'onprem.pipelines.summarizer': { 'onprem.pipelines.summarizer.Summarizer': ( 'pipelines.summarizer.html#summarizer',
                                                                                         'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.__init__': ( 'pipelines.summarizer.html#summarizer.__init__',
                                                                                                  'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._map_reduce': ( 'pipelines.summarizer.html#summarizer._map_reduce',
                                                                                                     'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._refine': ( 'pipelines.summarizer.html#summarizer._refine',
                                                                                                 'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize': ( 'pipelines.summarizer.html#summarizer.summarize',
                                                                                                   'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize_by_concept': ( 'pipelines.summarizer.html#summarizer.summarize_by_concept',
                                                                                                              'onprem/pipelines/summarizer.py')},
            'onprem.sk.base': { 'onprem.sk.base.decode_by_line': ('sk.base.html#decode_by_line', 'onprem/sk/base.py'),
                                'onprem.sk.base.detect_encoding': ('sk.base.html#detect_encoding', 'onprem/sk/base.py'),
                                'onprem.sk.base.detect_lang': ('sk.base.html#detect_lang', 'onprem/sk/base.py'),
                                'onprem.sk.base.get_random_colors': ('sk.base.html#get_random_colors', 'onprem/sk/base.py'),
                                'onprem.sk.base.is_chinese': ('sk.base.html#is_chinese', 'onprem/sk/base.py'),
                                'onprem.sk.base.is_nospace_lang': ('sk.base.html#is_nospace_lang', 'onprem/sk/base.py'),
                                'onprem.sk.base.split_chinese': ('sk.base.html#split_chinese', 'onprem/sk/base.py')},
            'onprem.sk.clf': { 'onprem.sk.clf.Classifier': ('sk.clf.html#classifier', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.__init__': ('sk.clf.html#classifier.__init__', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.create_model': ('sk.clf.html#classifier.create_model', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.evaluate': ('sk.clf.html#classifier.evaluate', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.fit': ('sk.clf.html#classifier.fit', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.grid_search': ('sk.clf.html#classifier.grid_search', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.load': ('sk.clf.html#classifier.load', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.load_texts_from_csv': ( 'sk.clf.html#classifier.load_texts_from_csv',
                                                                                 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.load_texts_from_folder': ( 'sk.clf.html#classifier.load_texts_from_folder',
                                                                                    'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.predict': ('sk.clf.html#classifier.predict', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.predict_proba': ('sk.clf.html#classifier.predict_proba', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.Classifier.save': ('sk.clf.html#classifier.save', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.NBSVM': ('sk.clf.html#nbsvm', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.NBSVM.__init__': ('sk.clf.html#nbsvm.__init__', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.NBSVM._fit_binary': ('sk.clf.html#nbsvm._fit_binary', 'onprem/sk/clf.py'),
                               'onprem.sk.clf.NBSVM.fit': ('sk.clf.html#nbsvm.fit', 'onprem/sk/clf.py')},
            'onprem.sk.tm': { 'onprem.sk.tm.TopicModel': ('sk.tm.html#topicmodel', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.__init__': ('sk.tm.html#topicmodel.__init__', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._check_build': ('sk.tm.html#topicmodel._check_build', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._check_model': ('sk.tm.html#topicmodel._check_model', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._check_recommender': ('sk.tm.html#topicmodel._check_recommender', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._check_scorer': ('sk.tm.html#topicmodel._check_scorer', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._harden_topics': ('sk.tm.html#topicmodel._harden_topics', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel._rank_documents': ('sk.tm.html#topicmodel._rank_documents', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.build': ('sk.tm.html#topicmodel.build', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.filter': ('sk.tm.html#topicmodel.filter', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_docs': ('sk.tm.html#topicmodel.get_docs', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_doctopics': ('sk.tm.html#topicmodel.get_doctopics', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_document_topic_distribution': ( 'sk.tm.html#topicmodel.get_document_topic_distribution',
                                                                                           'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_sorted_docs': ('sk.tm.html#topicmodel.get_sorted_docs', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_texts': ('sk.tm.html#topicmodel.get_texts', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_topics': ('sk.tm.html#topicmodel.get_topics', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.get_word_weights': ('sk.tm.html#topicmodel.get_word_weights', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.predict': ('sk.tm.html#topicmodel.predict', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.print_topics': ('sk.tm.html#topicmodel.print_topics', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.recommend': ('sk.tm.html#topicmodel.recommend', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.save': ('sk.tm.html#topicmodel.save', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.score': ('sk.tm.html#topicmodel.score', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.search': ('sk.tm.html#topicmodel.search', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.topics': ('sk.tm.html#topicmodel.topics', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.train': ('sk.tm.html#topicmodel.train', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.train_recommender': ('sk.tm.html#topicmodel.train_recommender', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.train_scorer': ('sk.tm.html#topicmodel.train_scorer', 'onprem/sk/tm.py'),
                              'onprem.sk.tm.TopicModel.visualize_documents': ( 'sk.tm.html#topicmodel.visualize_documents',
                                                                               'onprem/sk/tm.py')},
            'onprem.utils': { 'onprem.utils.SafeFormatter': ('utils.html#safeformatter', 'onprem/utils.py'),
                              'onprem.utils.SafeFormatter.__init__': ('utils.html#safeformatter.__init__', 'onprem/utils.py'),
                              'onprem.utils.SafeFormatter._replace_match': ('utils.html#safeformatter._replace_match', 'onprem/utils.py'),
                              'onprem.utils.SafeFormatter.format': ('utils.html#safeformatter.format', 'onprem/utils.py'),
                              'onprem.utils.SafeFormatter.parse': ('utils.html#safeformatter.parse', 'onprem/utils.py'),
                              'onprem.utils.batch_generator': ('utils.html#batch_generator', 'onprem/utils.py'),
                              'onprem.utils.batch_list': ('utils.html#batch_list', 'onprem/utils.py'),
                              'onprem.utils.contains_sentence': ('utils.html#contains_sentence', 'onprem/utils.py'),
                              'onprem.utils.df_to_md': ('utils.html#df_to_md', 'onprem/utils.py'),
                              'onprem.utils.download': ('utils.html#download', 'onprem/utils.py'),
                              'onprem.utils.filtered_generator': ('utils.html#filtered_generator', 'onprem/utils.py'),
                              'onprem.utils.format_string': ('utils.html#format_string', 'onprem/utils.py'),
                              'onprem.utils.get_datadir': ('utils.html#get_datadir', 'onprem/utils.py'),
                              'onprem.utils.get_models_dir': ('utils.html#get_models_dir', 'onprem/utils.py'),
                              'onprem.utils.get_template_vars': ('utils.html#get_template_vars', 'onprem/utils.py'),
                              'onprem.utils.get_webapp_dir': ('utils.html#get_webapp_dir', 'onprem/utils.py'),
                              'onprem.utils.html_to_df': ('utils.html#html_to_df', 'onprem/utils.py'),
                              'onprem.utils.md_to_df': ('utils.html#md_to_df', 'onprem/utils.py'),
                              'onprem.utils.remove_sentence': ('utils.html#remove_sentence', 'onprem/utils.py'),
                              'onprem.utils.segment': ('utils.html#segment', 'onprem/utils.py')}}}



"""A module to guide the output of an LLM using conditions and constraints"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/04_pipelines.guider.ipynb.

# %% auto 0
__all__ = ['Guider']

# %% ../../nbs/04_pipelines.guider.ipynb 3
from guidance import models

# %% ../../nbs/04_pipelines.guider.ipynb 4
class Guider:
    def __init__(
        self,
        llm,
        **kwargs,
    ):
        """
        Guider Constructor. Note that the `callback` and`callback_manager` parameters to `onprem.LLM` 
        are not currently utilized by the `Guider`.

        **Args:**

        - *llm*: An `onprem.LLM` object

        """
        self.llm = llm
        if not llm.is_llamacpp():
            raise ValueError(f'The suppplied LLM is {llm.model_url}, but the Guider currently only supports local, on-premises llama.cpp-based models.')


    def prompt(self, guidance_program: str or guidance._grammar.Join, echo=True):
        """
        A guided prompt. Input is a [Guidance program](https://github.com/guidance-ai/guidance) (`guidance>=0.1.0`) 
        that specifies constraints and conditions on the output for the prompt.

        **Args:**

        - *guidance_program*: A Guidance program (`>=0.1.0`) in the form a string or a `guidance._grammar.Join` object
        - *echo*: If True, echo highlighted output in Jupyter notebook. Set `echo=False` if running stand-alone script.

        **Returns:**

        - A dictionary with keys specified in the Guidance program and values containing the model outputs
        """
        model = models.LlamaCpp(self.llm.llm.client, echo=echo) 
        output = model + guidance_program
        return output.__dict__['_variables']



"""Token Detection module"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/05_hf.models.tokendetection.ipynb.

# %% auto 0
__all__ = ['TokenDetection']

# %% ../../../nbs/05_hf.models.tokendetection.ipynb 3
"""
Token Detection module
"""

import inspect
import os

import torch

from transformers import PreTrainedModel


class TokenDetection(PreTrainedModel):
    """
    Runs the replaced token detection training objective. This method was first proposed by the ELECTRA model.
    The method consists of a masked language model generator feeding data to a discriminator that determines
    which of the tokens are incorrect. More on this training objective can be found in the ELECTRA paper.
    """

    def __init__(self, generator, discriminator, tokenizer, weight=50.0):
        """
        Creates a new TokenDetection class.

        Args:
            generator: Generator model, must be a masked language model
            discriminator: Discriminator model, must be a model that can detect replaced tokens. Any model can
                           can be customized for this task. See ElectraForPretraining for more.
        """

        # Initialize model with discriminator config
        super().__init__(discriminator.config)

        self.generator = generator
        self.discriminator = discriminator

        # Tokenizer to save with generator and discriminator
        self.tokenizer = tokenizer

        # Discriminator weight
        self.weight = weight

        # Share embeddings if both models are the same type
        # Embeddings must be same size
        if self.generator.config.model_type == self.discriminator.config.model_type:
            self.discriminator.set_input_embeddings(self.generator.get_input_embeddings())

        # Set attention mask present flags
        self.gattention = "attention_mask" in inspect.signature(self.generator.forward).parameters
        self.dattention = "attention_mask" in inspect.signature(self.discriminator.forward).parameters

    # pylint: disable=E1101
    def forward(self, input_ids=None, labels=None, attention_mask=None, token_type_ids=None):
        """
        Runs a forward pass through the model. This method runs the masked language model then randomly samples
        the generated tokens and builds a binary classification problem for the discriminator (detecting if each token is correct).

        Args:
            input_ids: token ids
            labels: token labels
            attention_mask: attention mask
            token_type_ids: segment token indices

        Returns:
            (loss, generator outputs, discriminator outputs, discriminator labels)
        """

        # Copy input ids
        dinputs = input_ids.clone()

        # Run inputs through masked language model
        inputs = {"attention_mask": attention_mask} if self.gattention else {}
        goutputs = self.generator(input_ids, labels=labels, token_type_ids=token_type_ids, **inputs)

        # Get predictions
        preds = torch.softmax(goutputs[1], dim=-1)
        preds = preds.view(-1, self.config.vocab_size)

        tokens = torch.multinomial(preds, 1).view(-1)
        tokens = tokens.view(dinputs.shape[0], -1)

        # Labels have a -100 value to ignore loss from unchanged tokens
        mask = labels.ne(-100)

        # Replace the masked out tokens of the input with the generator predictions
        dinputs[mask] = tokens[mask]

        # Turn mask into new target labels - 1 (True) for corrupted, 0 otherwise.
        # If the prediction was correct, mark it as uncorrupted.
        correct = tokens == labels
        dlabels = mask.long()
        dlabels[correct] = 0

        # Run token classification, predict whether each token was corrupted
        inputs = {"attention_mask": attention_mask} if self.dattention else {}
        doutputs = self.discriminator(dinputs, labels=dlabels, token_type_ids=token_type_ids, **inputs)

        # Compute combined loss
        loss = goutputs[0] + self.weight * doutputs[0]
        return loss, goutputs[1], doutputs[1], dlabels

    def save_pretrained(self, output, state_dict=None, **kwargs):
        """
        Saves current model to output directory.

        Args:
            output: output directory
            state_dict: model state
            kwargs: additional keyword arguments
        """

        # Save combined model to support training from checkpoints
        super().save_pretrained(output, state_dict, **kwargs)

        # Save generator tokenizer and model
        gpath = os.path.join(output, "generator")
        self.tokenizer.save_pretrained(gpath)
        self.generator.save_pretrained(gpath)

        # Save discriminator tokenizer and model
        dpath = os.path.join(output, "discriminator")
        self.tokenizer.save_pretrained(dpath)
        self.discriminator.save_pretrained(dpath)




"""Wrappers for different approaches to text classification, including scikit-learn text classification, text classification with Hugging Face Transformers, and few-shot classification (via SetFit)."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/04_pipelines.classifier.ipynb.

# %% auto 0
__all__ = ['DATASET_TEXT', 'DATASET_LABEL', 'DEFAULT_SETFIT_MODEL', 'SMALL_SETFIT_MODEL', 'ClassifierBase', 'SKClassifier',
           'HFClassifier', 'FewShotClassifier']

# %% ../../nbs/04_pipelines.classifier.ipynb 3
from typing import List, Union

import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=DeprecationWarning)
    from setfit import SetFitModel, TrainingArguments, Trainer, sample_dataset

import numpy as np

# %% ../../nbs/04_pipelines.classifier.ipynb 4
from abc import ABC, abstractmethod

DATASET_TEXT = "text"
DATASET_LABEL = "label"

class ClassifierBase(ABC):

  
    @abstractmethod
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              max_steps:int=50,
              num_epochs:int=10,
              batch_size:int=32,
              metric='accuracy',
              callbacks=None,
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Additional keyword arguments are passed directly to `SetFit.TrainingArguments`.

        **Args:**

        - *X*: List of texts
        - *y*: List of integers representing labels
        - *max_steps*: If set to a positive number, the total number of training steps to perform. Overrides num_epochs. 
        - *num_epochs*: Number of epochs to train
        - *batch_size*: Batch size
        - *metric*: metric to use
        - *callbacks*:  A list of callbacks to customize the training loop.

        **Returns:**

        - None
        """
        pass

    @abstractmethod
    def save(self, save_path:str):
        """
        Save model to specified folder path, `save_path`
        """
        pass

    
    def get_explain_predictor(self):
        """
        Get predictor and tokenizer used for shap predictions.
        """
        return None, None

    
    def get_labels(self):
        return self.model.labels if hasattr(self.model, 'labels') else self.labels

        
    def sample_examples(self, X:list, y:list, num_samples:int=8,
                        text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Sample a dataset with `num_samples` per class
        """
        full_dataset = self.arrays2dataset(X, y, text_key=text_key, label_key=label_key)
                
        sample = sample_dataset(full_dataset, label_column=label_key, num_samples=num_samples)
        return sample.to_dict()[text_key], sample.to_dict()[label_key]
        

    
    def arrays2dataset(self, X:List[str], y:Union[List[int], List[str]], 
                       text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Convert train or test examples to HF dataset
        """
        from datasets import Dataset
        return Dataset.from_dict({text_key:X, label_key:y})


    def dataset2arrays(self, dataset, text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Convert a Hugging Face dataset to X, y arrays
        """
        return dataset.to_dict()['text'], dataset.to_dict()['label']
        
    def get_trainer(self):
        """
        Retrieves last trainer
        """
        if not self.trainer:
            raise ValueError('A trainer has not been created yet. You must first train a model on some labeled examples ' +\
                             'using the FewShotClassifier.train method.')
        return self.trainer

    
    def evaluate(self, X_eval:list, y_eval:list, print_report:bool=True, labels:list=[], **kwargs):
        """
        Evaluates labeled data using the trained model. 
        If `print_report` is True, prints classification report and returns nothing.
        Otherwise, returns and prints a dictionary of the results.
        Extra kwargs fed to `self.predict`.

        
        """
        labels = labels if labels else self.get_labels()
        labels = labels if labels else None
        
        from sklearn.metrics import classification_report
        y_pred= self.predict(X_eval, **kwargs)
        if isinstance(y_pred[0], str) and isinstance(y_eval[0], (np.integer, int)) and labels:
            y_eval = [labels[y] for y in y_eval]
        if isinstance(y_eval[0], str) and isinstance(y_pred[0], (np.integer, int)) and labels:
            y_pred = [labels[y] for y in y_pred]

        result = classification_report(y_eval, y_pred, 
                                       output_dict=not print_report,
                                       target_names = labels)
        if print_report:
            print(result)
            return
        else:
            import yaml
            #print(yaml.dump(result, allow_unicode=True, default_flow_style=False))
            return result


    def explain(self, X:list, labels:list=[]):
        """
        Explain the predictions on given examples in `X`. (Requires `shap` and `matplotlib` to be installed.)
        """
        X = [X] if isinstance(X, str) else X
        output_names = labels if labels else self.get_labels()
        output_names = output_names if output_names else None
        try:
            import shap
        except ImportError:
            raise ImportError('Please install the shap library: pip install shap')

        try:
            import matplotlib
        except ImportError:
            raise ImportError('Please install the matplotlib library: pip install matplotlib')

        f, tokenizer = self._get_explain_predictor()
        if f is None:
            raise NotImplementedError('Exlpanations are not currently supported for this model.')
        explainer = shap.Explainer(f, tokenizer, output_names=output_names)
        shap_values = explainer(X)
        shap.plots.text(shap_values)

# %% ../../nbs/04_pipelines.classifier.ipynb 10
class SKClassifier(ClassifierBase):
    def __init__(
        self,
        model_path=None,
        labels = [],
        **kwargs,
    ):
        """
        `SKClassifier` is a wrapper to scikit-learn text classifiation models.
        Extra kwargs are fed directly to `onprem.sk.clf.Classifier.create_model`.
        If no arguments are supplied, then a default Logistic Regression model is used.

        **Args:**

        - *model_path*: path to an already saved model file to be reloaded
        - *labels*: 
        - *labels*: list of strings intended to map label indices to string labels
        """

        from onprem.sk.clf import Classifier
        self.model = Classifier()
        self.labels = labels

        if model_path:
            self.model.load(model_path)
        else:
            # set defaults if necessary        
            if 'ctype' not in kwargs: 
                kwargs['ctype'] = 'sgdclassifier'
                kwargs['clf__loss'] = 'hinge'
                kwargs['clf__penalty'] = 'l2'
                kwargs['clf__alpha'] = 1e-3
                kwargs['clf__max_iter'] = 5
                kwargs['clf__tol'] = None
                kwargs['clf__random_state'] = 42
                kwargs['vec__token_pattern'] = '(?u)\\b\\w\\w+\\b'
                kwargs['use_tfidf'] = False

                # replace TF-IDF with this for slight accuracy boost
                # in relevant examples
                kwargs['vec__ngram_range'] = (1,3)
                kwargs['vec__binary'] = True
                kwargs['vec__max_features'] = 100000

            self.model.create_model(**kwargs) 

        
    def predict(self, X, **kwargs):
        """
        predict labels
        """
        labels = self.get_labels()
        preds = self.model.predict(X, **kwargs)
        preds = [preds] if not isinstance(preds, (list, np.ndarray)) else preds
        preds =  [labels[p] for p in preds] if labels else preds
        return preds[0] if len(preds) == 1 else preds

    
    def predict_proba(self, X, **kwargs):
        """
        predict label probabilities
        """
        return self.predict_proba(X, **kwargs)
    
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Additional keyword arguments are passed directly to `self.model.fit`.

        **Args:**

        - *X*: List of texts
        - *y*: List representing labels

        **Returns:**

        - None
        """

        self.model.fit(X, y)                                                                                                                   
  

    def save(self, filename:str):
        """
        Save model to specified `filename` (e.g., `/tmp/mymodel.gz`).
        Model saved as pickle file.
        To reload the model, supply `model_path` when instantiating`SKClassifier`.
        """
        self.model.save(filename)    

        

# %% ../../nbs/04_pipelines.classifier.ipynb 21
from ..hf import HFTrainer
from transformers import pipeline
import numpy as np
import os.path


class HFClassifier(ClassifierBase):
    def __init__(
        self,
        model_id_or_path:str='google/bert_uncased_L-2_H-128_A-2',
        device=None,
        labels=[],
        **kwargs,
    ):
        """
        `HFClassifier` can be used to train and run text Hugging Face transformer text classifiers.
                Additional keyword arguments are fed directly to `from_pretrained`.


        **Args:**

        - *model_id_or_path*: The Hugging Face model_id or path to model folder (e.g, path previously trained and saved model).
        - *device*: 'cuda' or 'cpu'
        - *labels*: list of strings intended to map label indices to string labels

        """
        self.model_id_or_path = model_id_or_path
        self.device=device
        self.model = None
        self.tokenizer = None
        self.labels = labels
        if os.path.isdir(self.model_id_or_path):
            from transformers import AutoTokenizer
            from transformers import AutoModelForSequenceClassification 
            self.model = AutoModelForSequenceClassification.from_pretrained(self.model_id_or_path)
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_id_or_path)

    def _tempsave(self):
        """
        Temporarily save model
        """
        import tempfile
        temp_dir = tempfile.TemporaryDirectory()
        if self.model:
            self.model.save_pretrained(temp_dir.name)
            self.tokenizer.save_pretrained(temp_dir.name)
        return temp_dir
    
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Extra kwargs are treated as arguments to `transformers.TrainingArguments`.

        **Args:**

        - *X*: List of texts
        - *y*: List of integers representing labels
        - *num_epochs*: Number of epochs to train
        - *batch_size*: Batch size
        - *metric*: metric to use
        - *callbacks*:  A list of callbacks to customize the training loop.

        **Returns:**

        - None
        """

        temp_dir = self._tempsave()
              
        # convert to HFTrainer format
        data = [{'text': x, 'label':y[i]} for i, x in enumerate(X)]

        # create a trainer
        trainer = HFTrainer()

        # train
        training_path = self.model_id_or_path if not self.model else temp_dir.name
        self.model, self.tokenizer = trainer(training_path,  data, **kwargs)

        # cleanup
        temp_dir.cleanup()
  

    def save(self, save_path:str):
        """
        Save model to specified folder path, `save_path`.
        To reload the model, supply path in `model_id_or_path` argument when
        instantiating`FewShotClassifier`.

        """
        self.model.save_pretrained(save_path)
        self.tokenizer.save_pretrained(save_path)


    def _get_pipeline(self):
        """
        Create transformers pipeline using current model
        """
        if not self.model:
            raise ValueError('There is not trained model yet. Please invoke the HFClassifier.train(...)')
        save_path = None
        if not os.path.isdir(self.model_id_or_path):
            temp_dir = self._tempsave()
            modelpath = temp_dir.name
        else:
            modelpath = self.model_id_or_path
        clf = pipeline('text-classification', model=modelpath, device=self.device)
        return clf
        
    def _predict(self, X:list, return_proba=True, max_length=512, **kwargs):
        """
        Predicts labels using `transformers.pipeline`
        """
        from operator import itemgetter
        X = [X] if isinstance(X, str) else X
        clf = self._get_pipeline(**kwargs)
        preds = []
        labels = self.get_labels()
        for example in clf(X, top_k=None, truncation=True, max_length=max_length):
            # unlike deprecated return_all_scores, top_k re-sorts
            example = sorted(example, key=itemgetter('label'))
            pred = [d['score'] for d in example]
            pred = np.argmax(pred) if not return_proba else pred
            pred = labels[pred] if labels else pred
            preds.append(pred)
        if len(preds) == 1:
            preds = preds[0]
        return preds

    def predict(self, X:list, max_length=512, **kwargs):
        """
        Predict labels. 
        Extra kwargs fed to Hugging Face transformers text-classification pipeline.
        """
        return self._predict(X, max_length=max_length, return_proba=False)
        
    def predict_proba(self, X:list, max_length=512, **wargs):
        """
        Predict labels.
        Extra kwargs fed to Hugging Face transformers text-classification pipeline.
        """
        return self._predict(X, max_length=max_length, return_proba=True)

    
    def _get_explain_predictor(self, device=None):
        return self._get_pipeline(), None
               


# %% ../../nbs/04_pipelines.classifier.ipynb 27
from sklearn.datasets import fetch_20newsgroups
from .classifier import HFClassifier

# %% ../../nbs/04_pipelines.classifier.ipynb 32
DEFAULT_SETFIT_MODEL = "sentence-transformers/paraphrase-mpnet-base-v2"
SMALL_SETFIT_MODEL = "sentence-transformers/all-MiniLM-L6-v2"

class FewShotClassifier(ClassifierBase):
    def __init__(
        self,
        model_id_or_path:str=DEFAULT_SETFIT_MODEL,
        use_smaller:bool=False,
        **kwargs,
    ):
        """
        `FewShotClassifier` can be used to train and run text classifiers. Currently based on SetFit.
                Additional keyword arguments are fed directly to `from_pretrained`.


        **Args:**

        - *model_id_or_path*: The Hugging Face model_id or path to model folder (e.g, path previously trained and saved model).
        - *use_smaller*:  If True, will use a smaller but performant model.

        """
        self.model_id_or_path = model_id_or_path
        if use_smaller and model_id_or_path != DEFAULT_SETFIT_MODEL:
            warnings.warn(f'Over-writing supplied model ({model_id_or_path}) with {SMALL_MODEL} because use_smaller=True.')
        self.model_id_or_path = SMALL_SETFIT_MODEL if use_smaller else self.model_id_or_path
        self.model = SetFitModel.from_pretrained(self.model_id_or_path, **kwargs)
        self.predict = self.model.predict
        self.predict_proba = self.model.predict_proba
        self.trainer = None # set in `FewShotClassifier.train`
        self.labels = [] # set in `FewShotClassifier.train`
            
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              num_epochs:int=10,
              batch_size:int=32,
              metric='accuracy',
              callbacks=None,
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Additional keyword arguments are passed directly to `SetFit.TrainingArguments`

        **Args:**

        - *X*: List of texts
        - *y*: List of integers representing labels
        - *num_epochs*: Number of epochs to train
        - *batch_size*: Batch size
        - *metric*: metric to use
        - *callbacks*:  A list of callbacks to customize the training loop.

        **Returns:**

        - None
        """

        # convert to HF dataset
        train_dataset = self.arrays2dataset(X, y, text_key='text', label_key='label')

        args = TrainingArguments(
                batch_size=batch_size,
                num_epochs=num_epochs,
                **kwargs
        )      

        trainer = Trainer(
                    model=self.model,
                    args=args,
                    metric=metric,
                    callbacks=callbacks,
                    train_dataset=train_dataset,
                    column_mapping={"text": "text", "label": "label"}
        )
        trainer.train()
      
        self.trainer = trainer
  

    def predict(self, X, **kwargs):
        """
        predict labels
        """
        return self.predict(X, **kwargs)


    def predict_proba(self, X, **kwargs):
        """
        predict label probabilities
        """
        return self.predict_proba(X, **kwargs)


    def save(self, save_path:str):
        """
        Save model to specified folder path, `save_path`.
        To reload the model, supply path in `model_id_or_path` argument when
        instantiating`FewShotClassifier`.

        """
        self.model.save_pretrained(save_path)        

    def _get_explain_predictor(self, device=None):
        def f(x):
            return self.predict_proba(x)

        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_id_or_path)
        return f, tokenizer
